# 
# Copyright (C) 2010-2012, 2013, 2014  Smithsonian Astrophysical Observatory
# 
# 
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# 

"""
Routines to support the fluximage tool.

We now display the "starting ..." message for each task when evaluating the
first chip/band/image/... rather than the previous behaviour of using a
barrier to write the message and then have the actual tasks depend on the
barrier. This is to avoid the case of having all the "starting" messages
be displayed (i.e. for all the obsids) before the actual processing starts.

The new approach means that the user gets a better idea of when the various
tasks are being run, but it's not technically correct since there's no
guarantee that the "first" task we create for a particular section is going
to get run first. For now I think it is worth it.
"""

import os
import sys
import tempfile
import shutil
import numpy as np
import subprocess as sbp

import cxcdm
import paramio
import pycrates as pcr
#import stk

import ciao_contrib.logger_wrapper as lw

from ciao_contrib.runtool import new_pfiles_environment
from ciao_contrib.stacklib import make_stackfile
from ciao_contrib.runtool import add_tool_history
from sherpa.utils import parallel_map

import ciao_contrib.ancillaryfiles as ancillary

import ciao_contrib._tools.fileio as fileio
import ciao_contrib._tools.utils as utils

import ciao_contrib._tools.run as run
from ciao_contrib._tools.run import \
    punlearn, dmcopy, dmimgcalc, dmimgcalc2, dmimgcalc_add, \
    dmhedit_key, dmhedit_file

# for now, export nothing
__all__ = ()

# Name of the background events file for HRC-I data.
HRCI_BG_EVTS = "HRC-I_bkg.fits"

lgr = lw.initialize_module_logger('_tools.fluximage')
v1 = lgr.verbose1
v2 = lgr.verbose2
v3 = lgr.verbose3
v4 = lgr.verbose4

def add_defargs(args, clobber, verbose):
    "Add clobber/verbse arguments to args list (modified *in place*)."

    if clobber:
        clstr = "yes"
    else:
        clstr = "no"

    args.append("clobber={}".format(clstr))
    args.append("verbose={}".format(verbose))

def add_energy_keywords(infile, enband, tmpdir="/tmp"):
    "Add the ENERGYLO/ENERGHI keywords to infile."

    tfile = tempfile.NamedTemporaryFile(dir=tmpdir, suffix=".keys")
    tfile.write("#add\n")
    tfile.write("COMMENT =Adding keywords for band={}\n".format(enband.bandlabel))
    tfile.write("ENERGYLO = {} / [keV] Min photon energy in the image\n".format(enband.elo))
    tfile.write("ENERGYHI = {} / [keV] Max photon energy in the image\n".format(enband.ehi))
    tfile.flush()

    try:
        dmhedit_file(infile, tfile.name, verbose=0)
    finally:
        tfile.close()

def copy_keywords(infile, outfile, keynames, tmpdir="/tmp"):
    """Copy the values of keynames from infile to outfile.

    If the keyword does not exist it is skipped.
    """

    if keynames == []:
        return

    vals = []
    cr = pcr.read_file(infile)
    try:
        for keyname in keynames:
            key = cr.get_key(keyname)
            if key == None:
                continue

            vals.append((keyname, key.value, key.unit, key.desc))

    finally:
        cr.get_dataset().snip()
        cr = None

    if vals == []:
        return

    tfile = tempfile.NamedTemporaryFile(dir=tmpdir, suffix=".keys")
    tfile.write("#add\n")
    tfile.write("COMMENT = Copying over keywords from " + infile)
    tfile.write("\n")

    for (name, value, unit, desc) in vals:

        # need to be careful with string values
        tfile.write("{} = ".format(name))
        if isinstance(value, basestring):
            if "/" in value:
                tfile.write("'{}'".format(value))
            else:
                tfile.write("\"'{}'\"".format(value))
        else:
            tfile.write("{}".format(value))
            
        if unit != "" or desc != "":
            tfile.write(" /")
            if unit != "":
                tfile.write(" [{}]".format(unit))

            if desc != "":
                tfile.write(" " + desc)

        tfile.write("\n")

    tfile.flush()

    try:
        dmhedit_file(outfile, tfile.name, verbose=0)
    finally:
        tfile.close()

def cleanup_files_task(filenames, msg=None):
    """Delete the files in the filenames array
    and, if msg is not None, display the message
    at verbose=2 level (after pre-pending the text
    'Cleaning up ')."""

    if msg != None:
        v2("Cleaning up {}".format(msg))

    v3("Deleting files: {}".format(filenames))
    for filename in filenames:
        os.unlink(filename)

def get_unique_vals(infile, colname, xrange=None, yrange=None):
    """Returns an array of the unique values for a column
    in the input file. An error is thrown if the column
    does not exist. If the column is empty then None is returned.

    If xrange is given then it is used as a filter on the
    x values in the file; [x={}:{}] is added to infile
    where xrange is assumed to be a pair of values.

    Similar for yrange, with the y column.

    """

    fname = infile
    if xrange != None:
        fname += "[x={}:{}]".format(*xrange)
    if yrange != None:
        fname += "[y={}:{}]".format(*yrange)

    return fileio.get_column_unique(fname, colname)

# Perhaps this should be included in the final output text
# rather than during processing?
#
def warn_unfilled_pixels((xlo, xhi), (ylo, yhi), (nx, ny), binsize):
    """Warn the user if the grid size leads to 'partially-filled' last
    row or column"""
    
    axes = 0
    if (xhi-xlo) % binsize != 0:
        axes |= 1
    if (yhi-ylo) % binsize != 0:
        axes |= 2

    if axes == 0:
        return
    
    if axes == 3:
        wstr = "row and column"
    elif axes == 2:
        wstr = "row"
    else:
        wstr = "column"

    v1("Fluxed image will underestimate the last {} since the exposure".format(wstr))
    v1("map is overestimated there. This is because the requested width and height")
    v1("is not a whole number of pixels.")

#################################################################
##
## File names
##
#################################################################
    
# In all these the head argument represents
# the location of the file (can be a directory or
# directory + string); it is expected to tbe outpath
# variable in the params dictionary.
# 
def name_asphist(head, num, blockname=False):
    """The name of the aspect histogram for chip/ccdid num.

    If blockname is True then [asphist] is appended.
    """

    o = "{}{}.asphist".format(head, num)
    if blockname:
        o += "[asphist]"
        
    return o
 
def name_instmap(head, num, enband):
    """The name of the instrument map for chip/ccdid num
    when evaluated with the given energy band.
    """

    return "{}{}_{}.instmap".format(head, num, enband.bandlabel)
 
def name_expmap(head, enband, num=None):
    """The name of the exposure map when evaluated for the
    given energy band.

    If num is not None then this is the chip/ccd number, otherwise
    the exposure map is for all chips/ccds.
    """

    if num == None:
        return "{}{}.expmap".format(head, enband.bandlabel)
    else:
        return "{}{}_{}.expmap".format(head, num, enband.bandlabel)

def name_expmap_thresh(head, enband):
    """The thresholded exposure map. Unlike name_expmap this is
    only for the per-observation exposure map (ie no support for the
    num argument).
    """

    return "{}{}_thresh.expmap".format(head, enband.bandlabel)

def name_image(head, enband):
    """The binned image for the given energy band.

    See also name_image_thresh() and name_fluxed().
    """

    if enband == None:
        raise ValueError("Internal error - name_image called with enband=None")

    return "{}{}.img".format(head, enband.bandlabel)

def name_image_thresh(head, enband):
    """The thresholded image for the given energy band.
    """

    return "{}{}_thresh.img".format(head, enband.bandlabel)

def name_fluxed(head, enband):
    """The fluxed (normalized by exposure map) image for the given
    energy band.
    """

    return "{}{}_flux.img".format(head, enband.bandlabel)

# This is actually only used by merge_all2, not by fluximage, so could
# move back to merge_all2. However, it requires routines in fluximage.py
# which depends on merging, so leave here for now to avoid cyclical
# dependencies.
# 
def fluximage_output(outdir, outhead, obsids, enbands,
                     threshold=False):
    """Return a dictionary listing all the output files that
    fluximage creates for the given inputs.
    """
    
    # TODO: SHOULD THIS BE SENT IN THE CLEANUP STATUS?

    def start():
        return dict([(enband.bandlabel,[]) for enband in enbands])
    
    # keys are energy band (as a string) and values are the list
    # of files in obsid order.
    #
    out = {
        'images': start(),
        'expmaps': start(),
        'fluxmaps': start(),
        }

    if threshold:
        out['image_thresh'] = start()
        out['expmap_thresh'] = start()

    # Going for simple over efficient.
    for enband in enbands:
        estr = enband.bandlabel

        for obsid in obsids:
            head = "{}{}{}_".format(outdir, outhead, obsid)

            out['images'][estr].append(name_image(head,enband))
            out['expmaps'][estr].append(name_expmap(head,enband))
            out['fluxmaps'][estr].append(name_fluxed(head,enband))

            if threshold:
                out['image_thresh'][estr].append(name_image_thresh(head,enband))
                out['expmap_thresh'][estr].append(name_expmap_thresh(head,enband))
            
    return out

    
##########################################################################
##########################################################################
##
## Start constructing routines to create exposure-corrected images
##
##########################################################################
##########################################################################

"""

The following was an attempt to avoid warning messages from
asphist with HRC data - e.g.

asphist (CIAO 4.4): WARNING: skipping 40 livetime correction records (from time: 61963173.370213 to time: 61963253.320216)

but it does not seem to help too much.

        efile = "{}[{}={}]".format(evtfile, filt, chip)
        timefilter = ""
        
        # if dtf file is not empty then add in a time filter
        # that matches the event file to avoid warning messages
        # like
        # asphist (CIAO 4.4): WARNING: skipping 40 livetime correction records (from time: 61963173.370213 to time: 61963253.320216)
        #
        # NOTE: not sure if correct/worth it
        #
        if dtffile != "":
            # note that unlike merging.reproject_events_task, we
            # use the first/last times from the event file itself
            # rather than from the aspect solutions
            # --> well, that didn't seem to fix things, so
            #
            #(t1, t2) = fileio.get_minmax_times(efile)
            #dtffile += "[time={}:{}]".format(t1, t2)
            
            atimes = []
            for afile in stk.build(asolobj.name):
                v3("Finding time range of aspect file: {}".format(afile))
                (t1, t2) = fileio.get_minmax_times(afile)
                atimes.append("{}:{}".format(t1, t2))
                v3(" >> {}".format(atimes[-1]))
                              
            if len(atimes) != 0:
                timefilter = "[time={}]".format(",".join(atimes))

"""

def run_asphist(outpath, asolobj, chip, evtfile, filt, dtffile,
                nbins, res_xy,
                message=None,
                verbose=0,
                clobber=False): 
    "Run asphist."

    if message != None:
        v1(message)
        
    with new_pfiles_environment(ardlib=False, copyuser=False):
        # punlearn("asphist")

        args = ["infile=" + asolobj.name,
                "outfile=" + name_asphist(outpath, chip),
                "evtfile={}[{}={}]".format(evtfile, filt, chip),
                "dtffile=" + dtffile,
                "max_bin={}".format(nbins),
                "res_xy={}".format(res_xy),
                "mode=h"]
        add_defargs(args, clobber, verbose)
        run.run("asphist", args)

def make_asphist(taskrunner, labelconv,
                 obs,
                 asolobj, chips, pixscale,
                 outpath,
                 parallel=True,
                 verbose=0,
                 clobber=False,
                 cleanup=True):
    """Calculate aspect histograms for the given chip(s) in this observation.
    asolobj is an AspectSolution object; pixscale is the final pixel size in
    arcsec and is used to scale the res_xy argument in asphist.
    """

    if obs.instrument == 'HRC':
        dtffile = obs.get_ancillary_('dtf')
        if dtffile == 'NONE':
            dtffile = ""
    else:
        dtffile = ""

    filt = obs.get_chipname()

    res_xy = 0.5
    if pixscale > 1.0:
        res_xy = pixscale / 2.0
    nbins = asolobj.nbins(xy=res_xy)
    
    tasks = []
    for chip in chips:
        task = labelconv("asphist-chip{}".format(chip))
        if chip == chips[0]:
            if len(chips) == 1:
                smsg = "Creating aspect histogram for obsid {}".format(obs.obsid)
            else:
                smsg = "Creating aspect histograms for obsid {}".format(obs.obsid)
        else:
            smsg = None

        taskrunner.add_task(task, [],
                            run_asphist, outpath, asolobj, chip,
                            obs.get_evtfile(),
                            filt,
                            dtffile,
                            nbins, res_xy,
                            message=smsg,
                            verbose=verbose,
                            clobber=clobber)
        tasks.append(task)

    etask = labelconv("asphist-end")
    taskrunner.add_barrier(etask, tasks)

    return etask

def reproject_bgnd(blanksky_map, outfile, asolfile, matchfile,
                   verbose="0",
                   clobber=False):
    "Reproject the background blank-sky events file."

    punlearn("reproject_events")
    args = ["infile=" + blanksky_map,
            "outfile=" + outfile,
            "aspect=" + asolfile,
            "match=" + matchfile,
            "random=0"]
    add_defargs(args, clobber, verbose)
    run.run("reproject_events", args)

def make_images_hrci(imgs,
                     bkgevents,
                     orig_bkgevts,
                     obsid,
                     outpath,
                     bin_filter,
                     tmpdir, ltable,
                     verbose="0"):
    """Subtract the HRC-I background (bkgevents) from the
    images (imgs array). The current support is such that
    imgs can only contain a single element, but leave in support
    for multiple files as the situation is in flux.

    orig_bkgevts is the original name of the background events
    file, so that it can be added to the image header using the
    BEVTFILE keyword.
    """

    if len(imgs) == 1:
        lbl = ""
    else:
        lbl = "s"
    imsg = "Subtracting HRC-I background{} for obsid {}".format(lbl, obsid)
    v1(imsg)

    bkg = tempfile.NamedTemporaryFile(dir=tmpdir, suffix=".bkg")
    dmcopy("{}[bin {}][opt type=i4]".format(bkgevents, bin_filter),
           bkg.name, clobber=True, verbose="0")

    with new_pfiles_environment(ardlib=False, copyuser=False):

        for img in imgs:
            hrcimg = tempfile.NamedTemporaryFile(dir=tmpdir, suffix=".bkg")

            dmimgcalc([img, bkg.name], hrcimg.name,
                      "img1-((img1_exposure/img2_exposure)*img2)",
                      lookupTab=ltable,
                      verbose=verbose, clobber=True,
                      tmpdir=tmpdir)

            # Edit the ONTIME/LIVETIME/EXPOSURE keywords of the
            # subtracted image so that they match the original image
            #
            copy_keywords(img, hrcimg.name, ["ONTIME", "LIVETIME", "EXPOSURE"])

            # Add in the name of the background file
            dmhedit_key(hrcimg.name, "BEVTFILE", orig_bkgevts,
                        comment="Background event file")
            
            # Use dmcopy rather than .copyfile to ensure that the
            # copy is recorded in the DM history
            # 
            dmcopy(hrcimg.name, img, clobber=True)

def create_event_image(infile, outfile,
                       tmpdir="/tmp",
                       verbose=0,
                       clobber=False):
    """This is basically

      dmcopy(infile, outfile)

    but it also removes the GTI block from the output file to
    (hopefully) make it easier to use the images downstream,
    in particular to get the correct EXPOSURE time in the
    exposure-corrected image.
    """

    v4("Binning up {} to {}".format(infile, outfile))
    tmpfile = tempfile.NamedTemporaryFile(dir=tmpdir, suffix='.img')
    dmcopy(infile, tmpfile.name, clobber=True, verbose=verbose)
    dmcopy(tmpfile.name + "[subspace - time]", outfile,
           clobber=clobber, verbose=verbose)

# TODO:
#   support providing a fov file to use
#
#   for HRC data, should be able to create energy-band independant images
#   (ie only one file rather than one per band).
#
def make_event_images(taskrunner, labelconv,
                      obs,
                      bkgevts,
                      asol,
                      grid,
                      enbands,
                      chips,
                      outpath,
                      lookuptab,
                      tmpdir="/tmp",
                      verbose="0",
                      clobber=False,
                      cleanup=True):
    """Create the binned count images. The return value is the list
    of file names that have been created.

    evtfile should include any filters given by the user and grid
    is the sky grid to use for the images. asol is a comma-separated
    string containing the asol file names.

    enbands is the set of energy bands with unique energy ranges
    (i.e. there are no repeat a:b pairs in this list of a:b:c values).
    """

    filename = obs.get_evtfile()
    if obs.instrument == "ACIS" and obs.grating != "NONE":
        filename += "[tg_m=0,Null]"

    out = []
    tasks = []
    for enband in enbands:
        v3("Creating an image with elo={} ehi={}".format(enband.elo, enband.ehi))
        ifile = "{}[bin {}]{}[opt type=i4]".format(filename, grid, enband.enfilterstr)
        ofile = name_image(outpath, enband)

        task = labelconv("evtbin-{}".format(enband.bandlabel))
        taskrunner.add_task(task, [], create_event_image, ifile, ofile,
                            tmpdir=tmpdir, clobber=clobber, verbose=verbose)
        out.append(ofile)

        etask = labelconv(ofile)
        taskrunner.add_task(etask, [task], add_energy_keywords,
                            ofile, enband, tmpdir=tmpdir)
        tasks.append(etask)

    etask = labelconv("evtbin-end")
    taskrunner.add_barrier(etask, tasks)
    if bkgevts == None:
        return etask
    
    btask = background_subtract_hrci(taskrunner, labelconv, [etask],
                                     obs,
                                     out,
                                     bkgevts,
                                     asol,
                                     grid,
                                     outpath,
                                     lookuptab,
                                     tmpdir=tmpdir,
                                     verbose=verbose,
                                     clobber=clobber,
                                     cleanup=cleanup
                                     )

    return btask

def reproject_bgnd_hrci(bkgevts, evtfile, obsid, asol, reproj,
                        tmpdir="/tmp/",
                        verbose=0,
                        clobber=False
                        ):

    bevt = blanksky_hrci(bkgevts, evtfile,
                         tmpdir, obsid, verbose)
    reproject_bgnd(bevt.name,
                   reproj,
                   asol, evtfile,
                   verbose=verbose,
                   clobber=clobber)

def background_subtract_hrci(taskrunner, labelconv, preconditions,
                             obs,
                             imgs,
                             bkgevts, asol,
                             grid,
                             outpath,
                             lookup_table,
                             tmpdir="/tmp/",
                             verbose="0",
                             clobber=False,
                             cleanup=True):
    """Background subtract the HRC-I counts data (the file names in the
    bimgs array).

    bkgevts is the path to the HRC-I background file (likely in
    the CALDB but a manual version could be used), 
    evtfile should include any filters given by the user, asol is
    a string containing a comma-separated list of asol files,
    and grid is the sky grid to use for the images.
    """

    reproj = outpath + HRCI_BG_EVTS

    stask = labelconv("reproject-hrci-bgnd")
    taskrunner.add_task(stask, preconditions,
                        reproject_bgnd_hrci,
                        bkgevts, 
                        obs.get_evtfile(),
                        obs.obsid, asol, reproj,
                        tmpdir=tmpdir, verbose=verbose, clobber=clobber)
                        
    # at the moment this is only called for one file so no need to make
    # it parallel.
    etask = labelconv("subtract-hrci-bgnd")
    taskrunner.add_task(etask, [stask],
                        make_images_hrci,
                        imgs,
                        reproj,
                        bkgevts,
                        obs.obsid,
                        outpath,
                        grid,
                        tmpdir,
                        lookup_table,
                        verbose=verbose)

    if cleanup:
        etask2 = labelconv("subtract-hrci-bgnd-cleanup")
        taskrunner.add_task(etask2, [etask], cleanup_files_task, [reproj],
                            "HRC-I background events file")
        return etask2
    
    else:
        return etask

def get_imap_nbins(maxsize, npixels):
    """Calculate the largest integer binsize that divides into npixels
    and is no larger than maxsize, and return the number of pixels
    of this length.

    If maxsize < 1, which is valid, then return npixels.
    """

    ms = np.ceil(maxsize).astype(np.int)
    xs = [i for i in range(ms, 0, -1) if npixels % i == 0]
    if xs == []:
        # We should not be able to get here and hit this situation, but
        # just in case
        raise ValueError("Unable to find any binsizes for npixels={} and max binsize={}".format(npixels, maxsize))

    return npixels / xs[0]

def instrument_map_hrc(obs, outpath, chips, energies, binval, units):
    """Return the arguments needed for the mkinstmap calls for HRC data.

    obs is a ciao_contrib._tools.obsinfo.ObsInfo object.
    energies is an array of energy bands; assumed to be unique.

    Return is (pixelgrid, args).
    """

    out = []
    if obs.detector == "HRC-I":
        detsubsys = "HRC-I"
        nbins = get_imap_nbins(binval, 16384)
        pixelgrid = "1:16384:#{0},1:16384:#{0}".format(nbins)
        v3("HRC-I instrument map will be {} pixels square.".format(nbins))

        if units == "time":
            detsubsys += ";IDEAL"

    elif obs.detector == "HRC-S":
        # We do not bin up the HRC-S instrument map because
        #   . 16456 mod 4096 is not 0
        #   . mkexpmap will create an empty output image with
        #     common binning factors on 64-bit machines
        #     TODO: is this still true in CIAO 4.6?
        #
        pixelgrid = "1:4096:#4096,1:16456:#16456"

    else:
        raise ValueError("Invalid DETNAM={} for HRC ({})".format(obs.detector, obs.get_evtfile()))

    badpix = obs.get_ancillary('bpix')
    if badpix == "CALDB":
        ardlib = None
    elif badpix == "NONE":
        ardlib = "AXAF_{}_BADPIX_FILE=NONE".format(obs.detector)
    else:
        ardlib = "AXAF_{}_BADPIX_FILE={}[BADPIX]".format(obs.detector, badpix)

    mask = obs.get_ancillary('mask')
    for j in chips:
        if obs.detector == "HRC-S":
            detsubsys = "HRC-S{}".format(j)
            if units == "time":
                detsubsys += ";IDEAL"
                
        # Was used in CIAO 4.5: 
        # aspfile = name_asphist(outpath, j, blockname=True)
        obsfile = obs.get_evtfile_no_dmfilter()

        if mask == "":
            maskfile = ""
        else:
            maskfile = "{}[MASK{}]".format(mask, j)

        for energy in energies:
            # Arguments are: outfile, maskfile, obsfile, detsubsys, monoenergy, weightfile, ardlib
            (enmono, wgtfile) = energy.instmap_options
            out.append(
                (name_instmap(outpath, j, energy),
                 maskfile,
            # aspfile,  - in CIAO 4.5 we could use the aspect file
                 obsfile,
                 detsubsys,
                 enmono,
                 wgtfile,
                 ardlib))

    return (pixelgrid, out)

def instrument_map_acis(obs, outpath, chips, energies, units):
    """Return the arguments needed for the mkinstmap calls for ACIS data.

    obs is a ciao_contrib._tools.obsinfo.ObsInfo object.
    energies is an array of energy bands; assumed to be unique.

    Return is (pixelgrid, args).

    In CIAO 4.6 we no longer return the pbkfile parameter.
    This may mean that very-old datasets will cause the tool to fail.
    """

    out = []
    pixelgrid = "1:1024:#1024,1:1024:#1024"

    maskfile = obs.get_ancillary('mask')
    badpix = obs.get_ancillary('bpix')

    for j in chips:
        detsubsys = "ACIS-{}".format(j)
        if units == "time":
            detsubsys += ";IDEAL"

        # Was used in CIAO 4.5: 
        # aspfile = name_asphist(outpath, j, blockname=True)
        obsfile = obs.get_evtfile_no_dmfilter()

        if badpix == "CALDB":
            ardlib = None
        elif badpix == "NONE":
            ardlib = "AXAF_ACIS{0}_BADPIX_FILE=NONE".format(j)
        else:
            ardlib = "AXAF_ACIS{0}_BADPIX_FILE={1}[BADPIX{0}]".format(j, badpix)

        for energy in energies:
            # Arguments are: outfile, maskfile, obsfile, detsubsys, monoenergy, weightfile, ardlib
            (enmono, wgtfile) = energy.instmap_options
            out.append(
                (name_instmap(outpath, j, energy),
                 maskfile,
            # aspfile,  - in CIAO 4.5 we could use the aspect file
                 obsfile,
                 detsubsys,
                 enmono,
                 wgtfile,
                 ardlib))

    return (pixelgrid, out)

def run_mkinstmap(outfile, mfile, obsfile, detsubsys, grating,
                  monoenergy, weightfile,
                  pixelgrid, mirror, dafile, units,
                  message=None,
                  verbose=0,
                  clobber=False,
                  ardlib=None):
    """Run mkinstmap in its own PFILES environment.

    In CIAO 4.5 the obsfile parameter was set to the aspect histogram
    for the chip - e.g. obsfile=xx/ciaox_0039_3.asphist[asphist] - but
    with the deprecation of the pbkfile parameter in CIAO 4.6 we
    now need to use the event file. This is not enforced by this
    routine (i.e. mkinstmap will error out if an invalid file is
    specified for obsfile).
    """

    if message != None:
        v1(message)
        
    with new_pfiles_environment(ardlib=False, copyuser=False):

        # punlearn("ardlib")
        if ardlib != None:
            run.run("pset", ["ardlib", ardlib])

        # punlearn("mkinstmap")
        args = [ "pixelgrid=" + pixelgrid,
                 "outfile=" + outfile,
                 "maskfile=" + mfile,
                 "obsfile=" + obsfile,
                 "detsubsys=" + detsubsys,
                 "grating=" + grating,
                 "spectrumfile={}".format(weightfile),
                 "monoenergy={}".format(monoenergy),
                 "mirror=" + mirror,
                 "dafile=" + dafile,
                 "mode=h"
                 ]
        add_defargs(args, clobber, verbose)
        run.run("mkinstmap", args)
        if units == "time":
            dmhedit_key(outfile, 'BUNIT', "", verbose=verbose)

def make_instrument_maps(taskrunner, labelconv, preconditions,
                         obs,
                         enbands,
                         chips,
                         units,
                         outpath,
                         binval,
                         parallel=True,
                         verbose="0",
                         clobber=False):
    """Create the instrument maps for the observation.

    enbands is the list of energy-band objects to process, but
    has been cleaned to ensure that they monochromatic energies
    are unique.

    units determins whether we want the default instrument
    map (count cm^2 photon^-1) - when units=default or area - 
    or time (no units).
    
    """
    
    if units == "default":
        mirror = "HRMA"
        dafile = "CALDB"

    elif units == "area":
        mirror = "HRMA"
        dafile = "CALDB"

    elif units == "time":
        mirror = "HRMA;area=1"
        dafile = "NONE"

    else:
        raise ValueError("Internal error: invalid setting units=" + units)

    if obs.instrument == "HRC":
        (pixelgrid, mkinstmap_args) = \
            instrument_map_hrc(obs,
                               outpath, chips, enbands,
                               binval,
                               units)

    elif obs.instrument == "ACIS":
        (pixelgrid, mkinstmap_args) = \
            instrument_map_acis(obs, outpath, chips, enbands, 
                               units)

    else:
        raise ValueError("Invalid INSTRUME={} keyword.".format(instrume))

    tasks = []
    for arg in mkinstmap_args:

        if arg == mkinstmap_args[0]:
            if len(mkinstmap_args) == 1:
                smsg = "Creating instrument map for obsid {}".format(obs.obsid)
            else:
                smsg = "Creating {} instrument maps for obsid {}".format(len(mkinstmap_args), obs.obsid)
        else:
            smsg = None
        
        (outfile, mfile, obsfile, detsubsys, monoenergy, wgtfile, ardlib) = arg

        # use the output file name as the label
        task = labelconv("imap-{}".format(outfile))

        taskrunner.add_task(task, preconditions, run_mkinstmap,
                            outfile, mfile, obsfile, detsubsys, obs.grating,
                            monoenergy, wgtfile,
                            pixelgrid, mirror, dafile, units,
                            message=smsg,
                            ardlib=ardlib, verbose=verbose,
                            clobber=clobber)
        tasks.append(task)

    etask = labelconv("imap-end")
    taskrunner.add_barrier(etask, tasks)
    return etask

def run_mkexpmap(outfile, instmap, asphist, matchfile, normalize,
                 message=None,
                 hackunits=True,
                 verbose=0,
                 clobber=False):
    "Create the given exposure map using a separate PFILES environment."

    if message != None:
        v1(message)
        
    with new_pfiles_environment(ardlib=False, copyuser=False):
        # punlearn("mkexpmap")

        # It is wasteful to re-calculate the grid each time,
        # but we do this to avoid having to pass information
        # between processes when running in parallel. Ideally
        # this grid would be imposed at the time the images are
        # created, so that it can be sent to further processes.
        #

        # We assume the same limits are used whatever the energy band or chip,
        # so we only need to call get_sky_limits once.
        #
        # if the grid does not 'fully fill' the image (i.e. if the last row/column
        # is part-filled due to the binning size) then we do not want to use get_sky_limits
        # here, since it assumes all rows/columns are fully filled. However, if we
        # do use a grid that does not fill the columns then you can get warning
        # messages from mkexpmap, along the lines of
        #
        #    ****** WARNING: The range [4470:5670:32,1810:3200:32] does not correspond to a
        #       whole number of pixels.  The following range will be used instead:
        #          [4470:5686:32]
        #       In the future, you should use the [min:max:#num] syntax to avoid this
        #       ambiguity.
        #
        # so we do in fact use the get_sky_limits approach and live with the fact
        # that this will result in over-correction for the image data.
        # 
        punlearn("get_sky_limits")
        run.run("get_sky_limits",
                ["image={}".format(matchfile),
                 "verbose=0",
                 "mode=h"]
                )
        xygrid = paramio.pgetstr("get_sky_limits", "xygrid")

        args = [ "outfile=" + outfile,
                 "instmapfile=" + instmap,
                 "asphistfile=" + asphist,
                 "xygrid=" + xygrid,
                 "normalize=" + normalize,
                 "useavgaspect=no",
                 "mode=h"
                 ]
        add_defargs(args, clobber, verbose)
        run.run("mkexpmap", args)

        # copy over a few keywords. In CIAO 4.4 mkexpmap does not copy over the
        # GRATING keyword, instead just setting it to NONE.
        copy_keywords(instmap, outfile, ["SPECTRUM", "WGTFILE", "ENERG_LO", "ENERG_HI", "GRATING"])

        if hackunits:
            dmhedit_key(outfile, 'BUNIT', 's', verbose=verbose)

def make_expmap_indiv(taskrunner, labelconv, preconditions,
                      outhead, obsid, enbands, chips, matchfile,
                      normalize="no",
                      hackunits=True,
                      parallel=True,
                      verbose="0",
                      clobber=False,
                      cleanup=True):
    """Create the per-chip, per obsid exposure maps.

    The energy bands are assumed to have unique monochromatic energies.

    If cleanup=True then the aspect histograms will be deleted
    once the exposure map has been created.
    """

    nruns = len(enbands) * len(chips)
    if nruns == 1:
        smsg = "Creating exposure map for obsid {}".format(obsid)
    else:
        smsg = "Creating {} exposure maps for obsid {}".format(nruns, obsid)

    tasks = []
    for j in chips:
        asphist = name_asphist(outhead, j, blockname=True)

        atasks = []
        for enband in enbands:
            outfile = name_expmap(outhead, enband, num=j)
            instmap = name_instmap(outhead, j, enband)

            task = labelconv("emap-{}".format(outfile))
            taskrunner.add_task(task, preconditions,
                                run_mkexpmap,
                                outfile, instmap, asphist, matchfile,
                                normalize,
                                message=smsg,
                                hackunits=hackunits,
                                verbose=verbose,
                                clobber=clobber
                                )

            smsg = None
            atasks.append(task)
            tasks.append(task)

        if cleanup:
            # Note: we do not make the emap-end depend on this task
            asphist = name_asphist(outhead, j, blockname=False)
            task2 = labelconv("asphist-{}-delete".format(asphist))
            taskrunner.add_task(task2, atasks,
                                cleanup_files_task, [asphist, instmap])
            
    etask = labelconv("emap-end")
    taskrunner.add_barrier(etask, tasks)
    return etask

# This originally used reproject_image with res=1 since it was
# not clear what the input grids would be. At present the images
# should be defined on exactly the same grids, so it is *just*
# an addition. So, we replace
#   reproject_iamge infile=@... matchfile=match outfile=out[EXPMAP]
#     method=average lookupTab=... resolution=1
# with
#   dmimgcalc infile=@... infile2= outfile=out[EXPMAP]
#     lookupTab=...
#
# Alternatives
#   - change resolution to 0 (significantly reducing run time)
#     in the reproject_image call
#   - use dmregrid2 instead (with either resolution=1 or 0)
#
# It looks like dmimgcalc and dmregrid2 with resolution=0 are
# similar in run times, but go with dmimgcalc for now.
#
# The code *could* check to see whether the grids agree, and
# fall back to reproject_image if they do not, but it is not
# worth it.
#
def combine_emaps(infiles, match, outfile, lookup_table, detnam,
                  message=None,
                  tmpdir="/tmp/",
                  verbose=0,
                  clobber=False):
    "Combine exposure maps, fix DETNAM and image units"

    if message != None:
        v1(message)

    with new_pfiles_environment(ardlib=False, copyuser=False):
        dmimgcalc_add(infiles,
                      outfile + "[EXPMAP]",
                      lookupTab=lookup_table,
                      clobber=clobber,
                      verbose=verbose,
                      tmpdir=tmpdir)

        # We could use dmhedit's filelist argument to just call the
        # tool once, with both DETNAM and BUNIT edits, but it is
        # not a huge time sink and this way we get a cleaner
        # history record.

        # restore the original DETNAM keyword
        dmhedit_key(outfile, 'DETNAM', detnam, verbose=verbose)

        # add the lost expmap BUNIT lost during dmimgcalc
        bunit = fileio.get_image_units(infiles[0])
        tunit = fileio.get_image_units(outfile)
        if bunit != tunit:
            dmhedit_key(outfile, 'BUNIT', bunit, verbose=verbose)

def single_expmap_chips(taskrunner, labelconv, preconditions,
                        outhead,
                        enbands,
                        chip,
                        verbose=0,
                        clobber=False,
                        cleanup=True):
    """There is only a single chip in this obsid so copy over
    to the output.
    """
    
    tasks = []
    delfiles = []
    for enband in enbands:
        infile = name_expmap(outhead, enband, num=chip)
        outfile = name_expmap(outhead, enband)

        task = labelconv("imap-copy-{}".format(infile))
        taskrunner.add_task(task, preconditions,
                            dmcopy, infile, outfile,
                            verbose=verbose, clobber=clobber)
        tasks.append(task)
        delfiles.append(infile)

    etask = labelconv("imap-combine-end")
    if cleanup:
        taskrunner.add_task(etask, tasks,
                            cleanup_files_task, delfiles)
    else:
        taskrunner.add_barrier(etask, tasks)
        
    return etask
    
def combine_expmap_chips(taskrunner, labelconv, preconditions,
                         outhead, enbands, obsid, chips, matchfile, detnam,
                         lookup_table, tmpdir,
                         parallel=True,
                         verbose=0,
                         clobber=False,
                         cleanup=True):
    """Combine per-chip exposure maps for each energy.

    The energy bands are assumed to have unique monochromatic energies.
    """

    # setup arguments to reproject images
    #
    fouthead = os.path.normpath(os.path.join(os.getcwd(), outhead))
    if outhead.endswith("/") and not fouthead.endswith("/"):
        # normpath removes the trailing / but we need it, so add it back in
        fouthead += "/"

    nchips = len(chips)
    nruns = len(enbands)
    if nruns == 1:
        smsg = "Combining {} exposure maps for obsid {}".format(nchips, obsid)
    else:
        smsg = "Combining {} exposure maps for {} bands (obsid {})".format(nchips, nruns, obsid)
    
    tasks = []
    for enband in enbands:

        # Ensure file names include the full path so that we can place the
        # stack file in any location.
        #
        expmaps = [name_expmap(fouthead, enband, num=j) for j in chips]
        task = labelconv("reproj-emap-{}".format(enband.bandlabel))
        taskrunner.add_task(task, preconditions,
                            combine_emaps,
                            expmaps, matchfile, name_expmap(outhead, enband),
                            lookup_table, detnam,
                            message=smsg,
                            verbose=verbose, clobber=clobber, tmpdir=tmpdir)

        smsg = None

        if cleanup:
            task2 = labelconv("reproj-emap-{}-cleanup".format(enband.bandlabel))
            taskrunner.add_task(task2, [task],
                                cleanup_files_task, expmaps)
            tasks.append(task2)
        else:
            tasks.append(task)

    etask = labelconv("reproj-emap-end")
    taskrunner.add_barrier(etask, tasks)
    return etask
    
def make_exposure_maps(taskrunner, labelconv, preconditions,
                       obs,
                       enbands,
                       chips,
                       outhead,
                       lookup_table,
                       normalize="no",
                       hackunits=False,
                       tmpdir="/tmp",
                       parallel=True,
                       verbose="0",
                       clobber=False,
                       cleanup=True
                       ):
    """Create exposure maps per chip and per band, and then for
    each band create a single exposure map (a copy if only one chip
    in use, otherwise adding up the individual exposure maps).

    The energy band list should have unique monochromatic energies.

    The hackunits field is set to True when the units of the
    exposure map should be changed from "cm**2 sec" to "s";
    it is an error to set hackunits=True and normalize="yes".

    If cleanup=True then the aspect solution, instrument maps,
    and per-chip exposure maps are deleted once they are finished
    with.
    """

    if hackunits and normalize == "yes":
        raise ValueError("Internal error: normalize=yes and hackuits=True")
    
    matchfile = name_image(outhead, enbands[0])

    # grid would be an input argument, set to the xygrid parameter setting
    #if grid.startswith("x="):
    #    xygrid = grid[2:].replace('y=', '')
    #else:
    #    raise ValueError("Expected grid argument to have form x=..,y=.. but sent {}".format(grid))

    etask = make_expmap_indiv(taskrunner, labelconv, preconditions,
                              outhead,
                              obs.obsid,
                              enbands,
                              chips,
                              matchfile,
                              parallel=parallel,
                              normalize=normalize,
                              hackunits=hackunits,
                              verbose=verbose,
                              clobber=clobber,
                              cleanup=cleanup)

    if len(chips) == 1:
        return single_expmap_chips(taskrunner, labelconv, [etask],
                                   outhead,
                                   enbands,
                                   chips[0],
                                   verbose=verbose,
                                   clobber=clobber,
                                   cleanup=cleanup)
    else:
        return combine_expmap_chips(taskrunner, labelconv, [etask],
                                    outhead, 
                                    enbands,
                                    obs.obsid,
                                    chips,
                                    matchfile,
                                    obs.detector,
                                    lookup_table,
                                    tmpdir,
                                    parallel=parallel,
                                    verbose=verbose,
                                    clobber=clobber,
                                    cleanup=cleanup)
                             
def run_imgthresh(infile, outfile, expmap, thresh,
                  message=None,
                  verbose=0,
                  clobber=False):
    "Runs dmimgthresh"

    if message != None:
        v1(message)
        
    with new_pfiles_environment(ardlib=False, copyuser=False):
        # punlearn("dmimgthresh")
        args = [ "infile=" + infile,
                 "outfile=" + outfile,
                 "expfile=" + expmap,
                 "cut=" + thresh,
                 "value=0.0",
                 "mode=h"]
        add_defargs(args, clobber, verbose)
        run.run("dmimgthresh", args)

# TODO: if use dmimgcalc_op with op=imgout=img1/img2 then will have
#   NaN rather than 0 for the areas with no exposure map.
#
def run_expcorr_image(imgfile, emapfile, outfile, lookup_table,
                      message=None,
                      verbose=0,
                      clobber=False):
    "Create exposure-corrected image"

    if message != None:
        v1(message)
        
    with new_pfiles_environment(ardlib=False, copyuser=False):
        v3("Creating fluxed image: {}".format(outfile))
        dmimgcalc2(imgfile, emapfile, outfile + "[PFLUX_IMAGE]",
                   "div", lookupTab=lookup_table,
                   verbose=verbose, clobber=clobber)

        # Copy over keywords originally from the instrument map
        copy_keywords(emapfile, outfile, ["SPECTRUM", "WGTFILE", "ENERG_LO", "ENERG_HI"])

        # Fix up the BUNIT field
        img_units = fileio.get_image_units(imgfile)
        emap_units = fileio.get_image_units(emapfile)
        v3("BUNIT cleanup: img={} emap={}".format(img_units, emap_units))

        # Some of these checks are not really necessary but left in in case
        # some of the tools change their BUNIT handling
        #
        if img_units == emap_units:
            fmap_units = fileio.get_image_units(outfile)
            if fmap_units != emap_units:
                v3("Cleaning out fluxed units from {} to ''".format(fmap_units))
                dmhedit_key(outfile, 'BUNIT', '')

        else:
            # Following http://heasarc.nasa.gov/docs/heasarc/ofwg/docs/general/ogip_93_001/ogip_93_001.html
            # but not trying to parse the existing units as too much effort
            # (although we expect to only have one of three cases so could just check)
            nkey = img_units
            if nkey != '' and emap_units != '':
                nkey += ' '

            if emap_units == 's':
                nkey += '/s'

            elif emap_units == 'cm**2':
                nkey += '/cm**2'

            elif emap_units == 'cm**2 s':
                nkey += '/cm**2 /s'

            elif emap_units == 'cm**2 sec':
                nkey += '/cm**2 /sec'

            else:
                nkey += '/(' + emap_units + ')'

            v3("New BUNIT={}".format(nkey))
            dmhedit_key(outfile, 'BUNIT', nkey)

def make_fluxed_images(taskrunner, labelconv, preconditions,
                       enbands,
                       obsid,
                       outhead,
                       lookup_table,
                       thresh=None,
                       parallel=True,
                       verbose="0",
                       clobber=False,
                       cleanup=True
                       ):
    "Create the fluxed images"

    thresh_args = []
    imcalc_args = []
    for enband in enbands:
        imgfile = name_image(outhead, enband)
        expmap  = name_expmap(outhead, enband)
        fluxmap = name_fluxed(outhead, enband)
            
        # thresh arguments are: imgfile, expmap, img_outfile, exp_outfile
        # imcalc arguments are: imgfile, expmap, outfile
        if utils.thresh_is_set(thresh):
            img_thrfile = name_image_thresh(outhead, enband)
            exp_thrfile = name_expmap_thresh(outhead, enband)
            thresh_args.append((imgfile, expmap, img_thrfile, exp_thrfile))
            imcalc_args.append((img_thrfile, exp_thrfile, fluxmap))
        else:
            imcalc_args.append((imgfile, expmap, fluxmap))

    nthresh = len(thresh_args)
    if nthresh > 0:
        smsg = "Thresholding data for obsid {}".format(obsid)

        tasks = []
        for (imgfile, expmap, img_thrfile, exp_thrfile) in thresh_args:
            itask = labelconv(img_thrfile)
            taskrunner.add_task(itask, preconditions,
                                run_imgthresh,
                                imgfile, img_thrfile, expmap, thresh,
                                message=smsg,
                                verbose=verbose, clobber=clobber)
            tasks.append(itask)
            smsg = None

            etask = labelconv(exp_thrfile)
            taskrunner.add_task(etask, preconditions,
                                run_imgthresh,
                                expmap, exp_thrfile, "", thresh,
                                message=smsg,
                                verbose=verbose, clobber=clobber)
            tasks.append(etask)

            if cleanup:
                ctask = labelconv("cleanup-{}-{}".format(img_thrfile,
                                                         exp_thrfile))
                taskrunner.add_task(ctask, [itask, etask],
                                    cleanup_files_task,
                                    [imgfile, expmap])

        preconditions = tasks

    ncalc = len(imcalc_args)
    if ncalc == 1:
        smsg = "Exposure-correcting image for obsid {}".format(obsid)
    else:
        smsg = "Exposure-correcting {} images for obsid".format(ncalc, obsid)

    tasks = []
    for (imgfile, expmap, fluxmap) in imcalc_args:
        task = labelconv(fluxmap)
        taskrunner.add_task(task, preconditions,
                            run_expcorr_image,
                            imgfile, expmap, fluxmap, lookup_table,
                            message=smsg,
                            clobber=clobber, verbose=verbose)
        smsg = None
        tasks.append(task)

    etask = labelconv("expcorr-end")
    taskrunner.add_barrier(etask, tasks)
    return etask

# TODO: no need to use crates for this; can just use normal I/O
def read_first_row(filename, colname):
    "Return the value of the first row of the given column."

    cr = pcr.read_file(filename + "[#row=1]")
    try:
        val = pcr.copy_colvals(cr, colname)[0]
    finally:
        cr.get_dataset().snip()
    return val

def find_blanksky_hrci(obsinfo, verbose=True, bgndmap=None):
    """Return the HRC-I background events file for the observation.

    The verbose flag determines whether to include informational
    messages at the verbose=1 level telling the user whether a
    background file was found or not.

    The bgndmap is intended for development/testing, and should
    contain a dictionary with keys being event file names (the same path as given
    to the script), and values being the full path to the
    HRC-I background file to use. If bgndmap is not None then it 
    is used instead of the CALDB lookup; missing files are treated
    as with a missing CALDB file, None is returned.

    There are two error conditions:
       - the HRC-I background files are not installed
       - there is no matching background file for the observation
       
    They are both signalled by (an optional screen message at verbose=1)
    and a return value of None. 
    """

    evtfile = obsinfo.get_evtfile()
    v3("Looking for HRC-I background file for {}".format(evtfile))

    if bgndmap == None:
        bname = _find_blanksky_hrci_caldb(evtfile, verbose=verbose)
    else:
        bname = _find_blanksky_hrci_manual(evtfile, bgndmap, verbose=verbose)
        
    if verbose and (bname != None):
        msg = "Background file for {} found: {}".format(evtfile, bname)
        v1(msg)

    return bname

def _find_blanksky_hrci_caldb(evtfile, verbose=True):
    "Find the HRC-I background file using a CALDB query"
    
    with new_pfiles_environment(ardlib=False, copyuser=False):
        v3("Querying CALDB for HRC-I background file")
        
        # hrc_bkgrnd_lookup displays the file name even with verbose=0 so we
        # explicitly hide it by piping to /dev/null. We also hide stderr since it
        # is unlikely that it adds anything useful to the error message we raise
        # below.
        #
        # punlearn("hrc_bkgrnd_lookup")
        devnull = open('/dev/null', 'w')    
        hbl_status = sbp.call(["hrc_bkgrnd_lookup",
                               evtfile,
                               "event"],
                              stdout=devnull, stderr=devnull)
        devnull.close()

        if hbl_status != 0:
            v3(" . hrc_bkgrnd_lookup call failed")
            if verbose:
                msg = "No HRC-I background file for {}, so skipping background subtraction.".format(evtfile)
                v1(msg)
                
            return None

        hrci_bkg = paramio.pgetstr("hrc_bkgrnd_lookup", "outfile")
        v3(" . hrc_bkgrnd_lookup found {}".format(hrci_bkg))
        
        # clean up file name for user consumption
        # os.path routines might be cleaner
        caldbdir = os.getenv("CALDB")
        clen = len(caldbdir)
        if caldbdir[-1] == "/":
            if hrci_bkg.startswith(caldbdir):
                bname = "$CALDB/" + hrci_bkg[clen:]
            else:
                bname = hrci_bkg
        else:
            if hrci_bkg.startswith(caldbdir) and hrci_bkg[clen] == "/":
                bname = "$CALDB" + hrci_bkg[clen:]
            else:
                bname = hrci_bkg

        if not os.path.exists(hrci_bkg):
            v3(" . hrc bkgrnd file not found on disk")
            if verbose:
                msg = "The background for {} is not in the CALDB, so skipping background subtraction.".format(evtfile)
                v1(msg)

            return None

        return hrci_bkg
    
def _find_blanksky_hrci_manual(evtfile, bgndmap, verbose=True):
    "Find the HRC-I background file using the given dictionary"

    key = evtfile
    v3("CALDB override: using a background map, with key={}".format(key))
    try:
        bname = bgndmap[key]
        v3(" -> found {}".format(bname))
    
    except KeyError:
        v3(" -> No match for key={} from file={} in bgndmap={}".format(key, evtfile, bgndmap))
        
        if verbose:
            msg = "WARNING: No HRC-I background file for {}, so skipping background subtraction.".format(evtfile)
            v1(msg)

        return None

    # since there may be DM filters
    try:
        pcr.read_file(bname + "[#row=1]")
        return bname

    except IOError:
        if not os.path.exists(bname):
            v3(" -> hrc bkgrnd file not found on disk!")
            if verbose:
                msg = "WARNING: The background for {} is not found, so skipping background subtraction.".format(evtfile)
                v1(msg)

            return None


def blanksky_hrci(caldbfile, infile, tmpdir, obsid, verbose):
    """Tailor the HRC-I background file in the CALDB (caldbfile)
    to match the given observation. Note that reprojection is not
    done here, just status filtering and header manipulation.

    The return value is the handle to the tailored events file;
    it will delete itself when it goes out of scope or the
    user explicitly closes it.
    """
    
    v1("Setting up the HRC-I background for obsid {}".format(obsid))

    bkg_stat = tempfile.NamedTemporaryFile(dir=tmpdir,
                                           suffix='.bkgevt')
    dmfilt = fileio.get_filter(caldbfile)
    if dmfilt == "":
        v3("Background file contains no DM filter, so can copy it")
        shutil.copyfile(caldbfile, bkg_stat.name)

    else:
        v3("Background file contains a DM filter ({}) so need to dmcopy it".format(dmfilt))
        dmcopy(caldbfile, bkg_stat.name, clobber=True, verbose="0")

    ofile = tempfile.NamedTemporaryFile(dir=tmpdir,
                                        suffix='.bkgevt')


    with new_pfiles_environment(ardlib=False, copyuser=False):

        # punlearn("dmmakepar")
        header = tempfile.NamedTemporaryFile(dir=tmpdir, suffix='.hdr')
        args = [ "input=" + infile,
                 "output=" + header.name,
                 "verbose=0",
                 "clobber=yes",
                 "mode=h"]
        run.run("dmmakepar", args)

        # We want the _PNT keywords, as well as
        #   OBJECT, OBS_ID, OBSERVER, TITLE, DS_IDENT
        #
        # DS_IDENT appears to be created automatically from the OBS_ID value,
        # but copy it across anyway.
        #
        # Other keywords, included to avoid warnings from merging:
        #   ASPTYPE
        #   DATAMODE  (bgnds have NEXT_IN_LINE it appears; is this a
        #              sensible change?)
        #   FOC_LEN
        #   BTIMDRFT  \
        #   BTIMNULL  | could probably just delete from all files
        #   BTIMRATE  /
        #
        # OBI_NUM appears to depend on whether the data has been reprocessed
        # or not
        #
        ts = ["ra", "dec", "roll"]
        wanted_terms = ["object", "obs_id", "observer", "title", "ds_ident", "seq_num"] \
            + [i + "_nom" for i in ts] \
            + [i + "_pnt" for i in ts] \
            + ["asptype", "datamode", "foc_len"] \
            + ["btimdrft", "btimnull", "btimrate"] \
            + ["obi_num"]
         
        pnt = tempfile.NamedTemporaryFile(dir=tmpdir,
                                          suffix='.par')
        h = open(header.name)
        p = open(pnt.name, "w")
        has_obi_num = False
        for line in h:
            term = line.split(",", 1)[0]
            if term in wanted_terms:
                p.write(line)
            if not has_obi_num and (term == "obi_num"):
                has_obi_num = True
                
        h.close()
        p.close()
        header.close()

        # update background header
        # punlearn("dmreadpar")
        args = [ "input=" + pnt.name,
                 "output=" + bkg_stat.name + "[EVENTS]",
                 "verbose=0",
                 "clobber=yes",
                 "mode=h"]
        run.run("dmreadpar", args)

        pnt.close()

        if not has_obi_num:
            try:
                run.run("dmhedit",
                        ["infile={}".format(bkg_stat.name),
                         "filelist=",
                         "operation=delete",
                         "key=OBI_NUM",
                         "verbose=0",
                         "mode=h"])
            except:
                pass

        # change status bits to match infile
        status = tempfile.NamedTemporaryFile(dir=tmpdir,
                                             suffix=".status") 
        args = [ "infile=" + infile,
                 "outfile=" + status.name,
                 "tool=dmcopy",
                 "action=get",
                 "clobber=yes"]
        run.run("dmhistory", args)

        statbits = str(read_first_row(status.name, "col2"))
        status.close()

        spos = statbits.find("[status")
        statbits = statbits[spos:statbits.find("]", spos)+1]

        dmcopy(bkg_stat.name + statbits, ofile.name,
               verbose=verbose, clobber=True)

    bkg_stat.close()
    return ofile

def clobber_checks(files, clobber=False):
    """If clobber is False then check that no output
    files already exist. If True, check that any that do
    exist can be written to.

    files is a dictionary where the values (can be a
    file name, array, or dictionary) are the file names.
    """

    v3("Clobber checks with clobber={}".format(clobber))
    fnames = []
    for (key, vals) in files.iteritems():
        v4("Checking for {} files.".format(key))
        if isinstance(vals, basestring):
            v4(" ... a single file")
            fnames.append(vals)

        elif isinstance(vals, dict):
            v4(" ... a dictionary")
            for val in vals.itervalues():
                fnames.extend(val)
            
        else:
            v4(" ... an array")
            fnames.extend(vals)
            
    for fname in fnames:
        fileio.outfile_clobber_checks(clobber, fname)
        
def get_output_filenames(outpath, enbands, chips, 
                         blanksky=False,
                         thresh=False):
    "Returns information on the files to be produced" 

    # TODO: split up exposure maps into per-observation and per-chip
    #
    asphists = [name_asphist(outpath, j) for j in chips]

    imgfiles = []
    threshimg = []
    threshemap = []
    instmaps = []
    expmaps = []
    fluxmaps = []

    for enband in enbands:
        instmaps.extend([name_instmap(outpath, j, enband) for j in chips])
        expmaps.extend([name_expmap(outpath, enband, num=j) for j in chips])
        expmaps.append(name_expmap(outpath, enband))

        imgfiles.append(name_image(outpath, enband))
        threshimg.append(name_image_thresh(outpath, enband))
        threshemap.append(name_expmap_thresh(outpath, enband))
        fluxmaps.append(name_fluxed(outpath, enband))

    # write dictionary of outputs
    outputs = {}
    outputs["images"] = utils.getUniqueSynset(imgfiles)
    outputs["asphists"] = utils.getUniqueSynset(asphists)
    outputs["instmaps"] = utils.getUniqueSynset(instmaps)
    outputs["expmaps"] = utils.getUniqueSynset(expmaps)
    outputs["fluxmaps"] = utils.getUniqueSynset(fluxmaps)

    # write dictionary of files that remain after a cleanup
    saved = {}
    saved["fluxmaps"] = outputs["fluxmaps"]

    # TODO: should the raw images not be output when threshold is on
    #       by default?
    if thresh:
        outputs["image_thresh"] = utils.getUniqueSynset(threshimg)
        outputs["expmap_thresh"] = utils.getUniqueSynset(threshemap)
        saved["image_thresh"] = outputs["image_thresh"]
        saved["expmap_thresh"] = outputs["expmap_thresh"]

    else:
        saved["images"] = outputs["images"]
        saved["expmaps"] = utils.getUniqueSynset([name_expmap(outpath, b)
                                                  for b in enbands])

    # list blank-sky maps
    if blanksky:
        outputs["blanksky"] = [outpath + HRCI_BG_EVTS]

    return (outputs, saved)


def add_history(outputs, cmdline_pars, toolname, toolversion,
                cleanup=True):
    """add a history entry to the file header of the output file using dmhistory
    and a note about the version of the fluximage script used.
    """

    v3("Updating file headers with history record.\n")

    add_hist = set()
    if cleanup:
        files = outputs[1]
    else:
        files = outputs[0]

        add_hist |= set(files["images"])
        add_hist |= set(files["asphists"])
        add_hist |= set(files["instmaps"])
        add_hist |= set(files["expmaps"])

        try:
            add_hist |= set(files["blanksky"])
        except KeyError:
            pass

    add_hist |= set(files["fluxmaps"])
    
    try:
        add_hist |= set(files["image_thresh"])
        add_hist |= set(files["expmap_thresh"])
    except KeyError:
        add_hist |= set(files["images"])
        add_hist |= set(files["expmaps"])

    # Assume that cmdline_pars contains all the parameter
    # values, even if not given directly on the command line
    # but read in from the parameter file.
    #
    add_tool_history(add_hist,
                     toolname,
                     cmdline_pars,
                     toolversion=toolversion)

SEPVAL = "\n     "

def clean_names(fnames):
    "Prepare array of file names for display"
    return SEPVAL + SEPVAL.join(fnames)

def add_files(label, fnames):
    "Return string identifying the file(s)."

    if len(fnames) == 1:
        x = " is"
    else:
        x = "s are"

    return " The {}{}:{}\n".format(label, x, clean_names(fnames))

# TODO: fix to deal with clipped/thresholded exposure maps
def print_output(outputs, cleanup=False):
    "Display the files that were created"

    if cleanup:
        files = outputs[1]
    else:
        files = outputs[0]

    created = ["", "The following files were created:", ""]

    if cleanup:
        try:
            created.append(add_files("clipped counts image", files["image_thresh"]))
        except KeyError:
            created.append(add_files("binned counts image", files["images"]))
    else:
        try:
            created.append(add_files("clipped counts image", files["image_thresh"]))
        except KeyError:
            pass

        created.append(add_files("binned counts image", files["images"]))
        
    if not cleanup:
        try:
            created.append(add_files("blank-sky events file", files["blanksky"]))
        except KeyError:
            pass

        created.extend([add_files("aspect histogram", files["asphists"]),
                        add_files("instrument map", files["instmaps"])])

    if cleanup:
        try:
            created.append(add_files("clipped exposure map", files["expmap_thresh"]))
        except KeyError:
            created.append(add_files("exposure map", files["expmaps"]))

    else:
        try:
            created.append(add_files("clipped exposure map", files["expmap_thresh"]))
        except KeyError:
            pass
        
        created.append(add_files("exposure map", files["expmaps"]))

    created.append(add_files("exposure-corrected image", files["fluxmaps"]))
    
    for c in created:
        v1(c)


##################################
##
## Combine the steps
##
##################################

def run_fluximage_tasks(taskrunner, labelconv,
                        obs,
                        bkgevtfile,
                        chips, 
                        asolobj, 
                        outpath, grid, binsize, enbands, units, thresh,
                        tmpdir="/tmp/",
                        verbose=0,
                        clobber=False,
                        cleanup=True,
                        parallel=False
                        ):
    """Run the various stages.

    The routine returns the task name that indicates that the
    fluxed images are available.
    """

    lookup_table = run.get_lookup_table("fluximage", pathfrom=__file__)
    if units == 'default':
        normflag = 'no'
    elif units == 'area':
        normflag = 'yes'
    elif units == 'time':
        normflag = 'no'
    else:
        raise ValueError("Internal error: invalid setting units={}".format(units))

    pixscale = utils.sky_to_arcsec(obs.instrument, binsize)
    
    asptask = make_asphist(taskrunner, labelconv,
                           obs,
                           asolobj,
                           chips,
                           pixscale,
                           outpath,
                           parallel=parallel,
                           verbose=verbose,
                           clobber=clobber
                           )

    imgtask = make_event_images(taskrunner, labelconv,
                                obs,
                                bkgevtfile,
                                asolobj.name,
                                grid,
                                enbands,
                                chips,
                                outpath,
                                lookup_table,
                                tmpdir=tmpdir,
                                verbose=verbose,
                                clobber=clobber,
                                cleanup=cleanup
                                )

    imaptask = make_instrument_maps(taskrunner, labelconv, [asptask],
                                    obs,
                                    enbands,
                                    chips,
                                    units,
                                    outpath,
                                    binsize,
                                    parallel=parallel,
                                    verbose=verbose,
                                    clobber=clobber
                                    )

    emaptask = make_exposure_maps(taskrunner, labelconv, [imgtask, imaptask],
                                  obs,
                                  enbands,
                                  chips,
                                  outpath,
                                  lookup_table,
                                  normalize=normflag,
                                  hackunits=(units == 'time'),
                                  tmpdir=tmpdir,
                                  parallel=parallel,
                                  verbose=verbose,
                                  clobber=clobber,
                                  cleanup=cleanup
                                  )

    fluxtask = make_fluxed_images(taskrunner, labelconv, [imgtask, emaptask],
                                  enbands,
                                  obs.obsid,
                                  outpath,
                                  lookup_table,
                                  thresh=thresh,
                                  parallel=parallel,
                                  verbose=verbose,
                                  clobber=clobber,
                                  cleanup=cleanup
                                  )

    return fluxtask

# End
