#
#  Copyright (C) 2010, 2011, 2012, 2013  Smithsonian Astrophysical Observatory
#
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Utility routines used by fluximage/merge_obs/...
"""

import os
import sys
import functools
import subprocess as sbp

import paramio
import stk

import ciao_contrib.logger_wrapper as lw

import ciao_contrib.runtool as rt
import coords.format

__all__ = (
    "ObsId",
    "make_obsid_from_headers",
    
    "set_verbosity",
    "print_version",
    "split_outroot",
    "to_number",
    "equal_elements",
    "getUniqueSynset",
    "thresh_is_set",
    
    "parse_range",
    "parse_xygrid",
    "parse_refpos",
    "sky_to_arcsec"
    )

lgr = lw.initialize_module_logger('_tools.utils')
v1 = lgr.verbose1
v3 = lgr.verbose3

@functools.total_ordering
class ObsId(object):
    """Represents an observation, as specified by
    an OBS_ID value and, optionally, "e1" or "e2"
    if an interleaved ACIS observation (corresponding
    to CYCLE=P/S).

    Unlike the header of the ACIS file, the cycle
    attribute will only be present if this file
    appears to be part of an interleaved-mode
    observation (either CYCLE=S or TIMEDELB!=0).

    Converting to a string uses the observation
    id followed by "e1" if a primary observation
    or "e2" if a secondary observation.

    Objects are judged to be equal if the obsid and
    cycle attributes are the same. Ordering is based on
    ObsId (lower values come first) and then primary
    before secondary.

    Use the make_obsid_from_headers() routine if you
    have a dictionary of headers.
    """

    def __init__(self, obsid, cycle=None):
        """Represent an ObsId: obsid and cycle should be the contents
        of the OBS_ID and CYCLE header keywords.

        Fails if obsid is 'merged' (case insensitive). Invalid
        CYCLE values (not 'P' or 'S') are ignored; in this case
        a warning is displayed at the verbose=1 level.
        """

        v3("ObsId sent obsid={} cycle={}".format(obsid, cycle))
        if obsid.lower() == "merged":
            raise ValueError("The OBS_ID value is '{}'".format(obsid))

        if cycle not in [None, 'P', 'S']:
            v1("WARNING: ObsId {} has unrecognized CYCLE={}, assuming not interleaved.".format(obsid))
            cycle = None
            
        self._obsid = obsid
        self._cycle = cycle

    @property
    def obsid(self):
        "The ObsId value."
        return self._obsid

    @property
    def cycle(self):
        "The Cycle value (None, 'S', or 'P')."
        return self._cycle

    def __str__(self):
        out = "{}".format(self.obsid)
        if self.cycle == 'P':
            out += 'e1'
        elif self.cycle == 'S':
            out += 'e2'
        return out

    def __repr__(self):
        "This is only for debugging rather than being valid Python"
        return "ObsId({},cycle={})".format(self.obsid,self.cycle)

    # no checks to avoid attribute errors if used with an invalid rhs
    def __eq__(self, rhs):
        return (self.obsid, self.cycle) == (rhs.obsid, rhs.cycle)

    def __ne__(self, rhs):
        return not self.__eq__(rhs)
    
    # Fortunately 'P' and 'S' have the ordering we want and we assume that
    # we do not end up comparing the same interleaved-mode ObsId where
    # one of them does not have information on its cycle.
    #
    def __lt__(self, rhs):
        return (self.obsid, self.cycle) < (rhs.obsid, rhs.cycle)

    # So we can use this as a key in a dictionary
    def __hash__(self):
        return hash((self.obsid, self.cycle))

def make_obsid_from_headers(headers, infile=None):
    """Create an ObsId object from the headers, which should
    accept dictionary-like access for the OBS_ID, CYCLE, and
    TIMEDELB keywords. The TIMEDELB keyword is used to determine
    whether this is an interleaved mode observatioin, when > 0,
    or not. The only required keyword is OBS_ID.

    This routine errors with a ValueError if OBS_ID indicates
    a merged observation.

    The infile argument is only used in error messages.
    """

    if infile != None:
        v3("Looking for ObsId from headers of {}.".format(infile))
    else:
        v3("Looking for ObsId from headers.")
        
    try:
        obsid = headers['OBS_ID']
    except KeyError:
        emsg = "The OBS_ID value is missing"
        if infile == None:
            emsg += "."
        else:
            emsg += " in {}.".format(infile)
        raise KeyError(emsg)
    except TypeError:
        raise TypeError("The input must accept string indexing.")

    if obsid.lower() == "merged":
        emsg = "The OBS_ID value is '{}'".format(obsid)
        if infile == None:
            emsg += "."
        else:
            emsg += " in {}.".format(infile)
        raise ValueError(emsg)

    try:
        cycle = headers['CYCLE']
        timedelb = headers['TIMEDELB']
        try:
            timedelb = float(timedelb)
        except ValueError:
            v3("ObsId: obsid={} cycle={} unable to convert timedelb={}".format(obsid, cycle, timedelb))
            cycle = None

        if timedelb > 0:
            v3("TIMEDEL={} is > 0 so assuming interleaved.".format(timedelb))
        else:
            v3("TIMEDEL={} is <= 0 so assuming non-interleaved.".format(timedelb))
            cycle = None
            
    except KeyError:
        v3("ObsId: obsid={}; does not contain CYCLE or TIMEDELB.".format(obsid))
        cycle = None
        
    if cycle not in [None, 'P', 'S']:
        v1("WARNING: ObsId {} has unrecognized CYCLE={}, assuming not interleaved.".format(obsid))
        cycle = None

    v3("ObsId: obsid={} cycle={}".format(obsid, cycle))
    return ObsId(obsid, cycle=cycle)

def print_version(toolname, version):
    """Print the name and version of the script"""

    v1("Running {}".format(toolname))
    v1("Version: {}\n".format(version))

def split_outroot(outroot):
    """Given the user's outroot, split it into
       (dirname, headname)
    where either component can be empty.

    If headname is not empty then we ensure it ends with
    '_'.
    """
    
    # split up the user's outroot into a directory name,
    # with the trailing /, and remaining text (which will
    # have a trailing _ added to it if one does not exist).
    #
    (dirname, head) = os.path.split(outroot)
    if dirname == "":
        outdir = ""
    else:
        outdir = dirname + "/"

    if head == '' or head.endswith('_'):
        outhead = head
    else:
        outhead = head + "_"

    v3("Splitting outroot parameter: {} to dirname={} head={}".format(outroot, outdir, outhead))
    return (outdir, outhead)

def to_number(inval):
    """Convert a string-formatted number into an integer or float.  The input is assumed
    to have been created by paramio.pget, so is assumed to either be a valid number of INDEF.

    INDEF is converted to None, '1.0' to 1 (integer), and '1.2' to 1.2 (float).
    Any invalid input will throw a ValueError.
    """

    if inval.lower() == "indef":
        return None
    else:
        try:
            out = float(inval)
        except ValueError:
            raise ValueError("Unable to convert '{0}' to a number.".format(inval))

    if int(out) == out:
        return int(out)
    else:
        return out

def equal_elements(ls):
    """test to see if all elements in a list are equal"""
    if ls:
        i = iter(ls)
        element = i.next()
        for value in i:
            if element != value:
                return False
    return True

def getUniqueSynset(ls):
    """get unique entries, while preserving order."""

    s = {}
    return [s.setdefault(u, u) for u in ls if u not in s]

def thresh_is_set(val):
    """Returns true if val is not one of
        None
        empty string
        none (case insensitive)
        0 or 0% or :
    """

    if val == None:
        return False
    
    return val.lower() not in ["", "none", "0", "0%", ":"]

def parse_range(rstr):
    """Convert a:b:#c or a:b:c into a tuple
    (lo, hi, binsize, nbins).
    """

    toks = rstr.split(":")
    if len(toks) != 3:
        raise ValueError("Expected a:b:c or a:b:#c, not {}".format(rstr))

    toks = [tok.strip() for tok in toks]
    lo = to_number(toks[0])
    hi = to_number(toks[1])
    if toks[2][0] == '#':
        nbins = to_number(toks[2][1:])
        if nbins <= 0:
            raise ValueError("Number of bins must be > 0 in {}".format(rstr))
        
        binsize = (hi - lo) * 1.0 / nbins
        
    else:
        binsize = to_number(toks[2])
        if binsize <= 0:
            raise ValueError("The binsize must be > 0 in {}".format(rstr))

        nbins = (hi - lo) * 1.0 / binsize

    return (lo, hi, binsize, nbins)

def parse_xygrid(gstr):
    """Given the xygrid parameter value, return the
    grid specification. We first check to see if there is
    a file that matches this value, and run get-sky_limits
    on it. If this fails we return the input string.

    The return value is actually
      (grid, binval, xrng, yrng, (nx,ny))

    where we use the minimum bin size if the two axes have
    a different value. There is a validation check on the
    string value. xrng and yrng are tuples containing the lo/hi
    values for each axes, and nx/ny are the number of bins
    in the output image for each axis.
    """

    v3("Verifying xygrid={}".format(gstr))

    with rt.new_pfiles_environment(ardlib=False, copyuser=False):
        devnull = open('/dev/null', 'w')
        sbp.call(["get_sky_limits", "image={}".format(gstr)],
                 stdout=devnull, stderr=devnull)
        devnull.close()
        xygrid = paramio.pgetstr("get_sky_limits", "xygrid")

    if xygrid == "":
        xygrid = gstr

    xystk = xygrid.split(",")
    if len(xystk) != 2:
        raise ValueError("xygrid should be a filename or 2 ranges separated by a comma, not '{}'!".format(xygrid))

    # Note: lose some info if there is an error
    try:
        xrng = parse_range(xystk[0])
        yrng = parse_range(xystk[1])

    except ValueError as ve:
        raise ValueError("xygrid should be a filename or xlo:xhi:#nx or dx,ylo:yhi:#ny or dy, not '{}'!".format(xygrid))

    xbinsize = xrng[2]
    ybinsize = yrng[2]

    # We pick the smaller binsize (expect them to be the same)
    if xbinsize != ybinsize:
        v1("WARNING: X and Y axis bin sizes are different ({} vs {}); using the smaller.".format(xbinsize, ybinsize))
        binsize = min(xbinsize, ybinsize)
    else:
        binsize = xbinsize

    if int(binsize) == binsize:
        binsize = int(binsize)

    return ("x={},y={}".format(xystk[0], xystk[1]), binsize,
            (xrng[0], xrng[1]), (yrng[0], yrng[1]), (xrng[3], yrng[3]))

# TODO: need to deal with a file name that includes a DM filter
def parse_refpos(refpos):
    """Given a reference position from a user - a string that can be
    empty, a file name, or ra/dec values, return the tuple

      (ra, dec, filename)

    where the values are None if not relevant, otherwise are
    numbers or a string.

    If no values are given then returns None.
    """

    if refpos.strip() == "":
        return None

    elif os.path.isfile(refpos):
        v3("Reference position is a file.")
        return (None, None, refpos)

    v3("Extracting reference position from " + refpos)

    # Assume that no spaces are used to separate out
    # sexagesimal formats, so that we can treat the
    # argument as a stack
    #
    coord = stk.build(refpos)
    if len(coord) != 2:
        raise ValueError("Unable to parse {} as a ra and dec value.".format(refpos))

    ra = coords.format.ra2deg(coord[0])
    dec = coords.format.dec2deg(coord[1])

    return (ra, dec, None)

def sky_to_arcsec(instrume, pixsize):
    """Given a distance in SKY pixels, convert to arc seconds.
    The instrume value should be ACIS or HRC.
    """

    if instrume == 'ACIS':
        return pixsize * 0.492
    elif instrume == 'HRC':
        return pixsize * 0.1318
    else:
        raise ValueError("Invalid instrume={} argument, expected ACIS or HRC.".format(instrume))
    
# End
