#
#  Copyright (C) 2010, 2011, 2012  Smithsonian Astrophysical Observatory
#
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Chandra CALDB routines.
"""

import os
import copy
import urllib2
import socket
from HTMLParser import HTMLParser
import time

import pycrates

__all__ = (
    "check_caldb_version",
    "get_caldb_dir",
    "get_caldb_releases",
    "get_caldb_installed",
    "get_caldb_installed_version",
    "get_caldb_installed_date"
    )

release_notes_url = "http://cxc.harvard.edu/caldb/downloads/releasenotes.html"

# CIAO 4.4 Crates workaround to ensure data is closed/memory reclaimed.
def _ciao44_crates_fix(cr):
    cr.get_dataset().snip()
    
def todate(txt):
    """Return a time object for the date format used
    by the CALDB tables. If txt is None then None is
    returned; if the text can not be converted then it
    is returned."""

    if txt == None:
        return None
    else:
        try:
            return time.strptime(txt+"UTC", "%Y-%m-%dT%H:%M:%S%Z")
        except ValueError:
            return txt

# The current version has the main data in a td with class=mainbar
# but this is hopefully going to change RSN for a div
#
class CALDBReleaseParser(HTMLParser):
    """Parse the CALDB release-notes HTML page for
    release information on the CALDB.

    The use is, where h is a string containing the HTML to
    parse,

      p = CALDBReleaseParser()
      p.feed(h)
      p.close()

    and then the p.releases field is a dictionary where the keys are
    the various CALDB release types (e.g. CALDB, SDP, CIAO, and L3)
    and the values are tuples of (version number, date) where the
    version number is a string and the date is a time object.

    An IOError is thrown if the result is empty (at least for the
    CALDB release column). An IOError is also returned if the
    page can not be parsed (e.g. it does not match the expected
    contents); in this case the error messages are not 'user friendly'
    since they reveal the internal state of the object to make
    debugging easier.
    """

    state = "need-table"
    open_mode = {
        #"need-main": { "tag": "td", "attribute": ("class", "mainbar"), "newstate": "need-table" },
        "need-table": { "tag": "table", "attribute": ("id", "caldbver"), "newstate": "check-header" },
        }

    close_mode = {
        "check-header": { "tag": "tr", "newstate": "store-data" },
        "store-data": { "tag": "table", "newstate": "finished" }
        }

    store = []
    row_store = None
    item_store = None

    def close(self):
        HTMLParser.close(self)
        if self.state != "finished":
            raise IOError("incorrectly-nested tables; state={0}".format(self.state))

        st = self.store
        self.releases = {
            "CALDB": [(s[0], s[1]) for s in st if s[1] != None],
            "SDP":   [(s[0], s[2]) for s in st if s[2] != None],
            "CIAO":  [(s[0], s[3]) for s in st if s[3] != None],
            "L3":    [(s[0], s[4]) for s in st if s[4] != None],
            }

        if len(self.releases["CALDB"]) == 0:
            raise IOError("No CALDB release information found!")

    def handle_starttag(self, tag, attrs):
        if self.state == "store-data":
            # Could do the storing via a pseudo state machine as we do with
            # finding the table, but just hard code the logic for now
            #
            if self.row_store == None:
                if tag == "tr":
                    self.row_store = []
                else:
                    raise IOError("A new row has started with the tag: {0}".format(tag))
            if tag == "td":
                if self.item_store == None:
                    self.item_store = ""
                else:
                    raise IOError("A new item has started but the item_store=[{0}]".format(self.item_store))
            return

        if self.state not in self.open_mode:
            return

        tbl = self.open_mode[self.state]

        if tag != tbl["tag"] or \
                ("attribute" in tbl and tbl["attribute"] not in attrs):
            return

        if "newstate" in tbl:
            self.state = tbl["newstate"]

    def handle_endtag(self, tag):
        if self.state == "store-data":
            if tag == "td":
                item = self.item_store.strip()
                if item.lower() in ["n/a", "not released publicly"]:
                    self.row_store.append(None)
                else:
                    self.row_store.append(item)
                self.item_store = None

            elif tag == "tr":
                r = self.row_store
                if len(r) != 5:
                    raise IOError("Unable to parse row: {0}".format(r))
                self.store.append((r[0],
                                   todate(r[1]),
                                   todate(r[2]),
                                   todate(r[3]),
                                   todate(r[4])))

                #self.store.append(tuple(self.row_store))
                self.row_store = None

        if self.state not in self.close_mode:
            return

        tbl = self.close_mode[self.state]
        if tag != tbl["tag"]:
            return

        self.state = tbl["newstate"]

    def handle_data(self, data):
        if self.state == "store-data" and self.item_store != None:
            ds = data.strip()
            if ds != None:
                if self.item_store == "":
                    self.item_store = ds
                else:
                    self.item_store += " {0}".format(ds)

def get_caldb_releases(timeout=None):
    """Return a dictionary containing the available CALDB releases,
    where the keys are the names "CALDB", "SDP", "CIAO" or "L3",
    and the values are arrays of (version-string, date) tuples. Each list
    is unique since the Chandra CALDB can be updated for a sub-set
    of these systems. There is no guarantee that the lists are
    in descending order of time or version number.
    
    The version-string is "1.2" or "4.2.2" and the date is a time
    object.

    This routine will only work if the computer is on-line and able to
    access the Chandra CALDB pages at

      http://cxc.harvard.edu/caldb/

    The timeout parameter is used to set the timeout used to
    access the CALDB web site; if set to None then the global
    default timeout will be used, otherwise it should be a value
    in seconds.
    """

    try:
        if timeout == None:
            h = urllib2.urlopen(release_notes_url).read()
        else:
            h = urllib2.urlopen(release_notes_url, timeout=timeout).read()

    except urllib2.URLError, ue:
        # Probably excessive attempt to make a "nice" error message
        #
        if hasattr(ue, "reason"):
            if hasattr(ue.reason, "errno") and ue.reason.errno == socket.EAI_NONAME:
                raise IOError("Unable to reach the CALDB site - is the network down?")
            else:
                raise IOError("Unable to reach the CALDB site - {0}".format(ue.reason))

        elif hasattr(ue, "getcode"):
            cval = ue.getcode()
            if cval == 404:
                raise IOError("The CALDB site appears to be unreachable.")
            else:
                raise IOError("The CALDB site returned {0}".format(ue))

        else:
            raise IOError("Unable to access the CALDB site - {0}".format(ue)) 

    try:
        p = CALDBReleaseParser()
        p.feed(h)
        p.close()
    except IOError, ie:
        raise IOError("Unable to parse the CALDB release table.")

    # use a deep copy so that the parser can be cleaned up
    # in case there's a lot of state squirreled away
    # (although have not checked that it actually matters)
    return copy.deepcopy(p.releases)

def get_caldb_dir():
    """Returns the location of the CALDB, as gicen by the CALDB environment
    variable, or throws an IOError if it is not set or can not be found."""

    caldb = os.getenv("CALDB")
    if caldb == None:
        raise IOError("CALDB environment variable is not defined!")
    elif not os.path.isdir(caldb):
        raise IOError("CALDB directory does not exist: {0}".format(caldb))
    else:
        return caldb

def get_caldb_installed(caldb=None):
    """Return a tuple containing the version (as a string) and
    date (as a date object) of the installed Chandra CALDB.

    If caldb is None then the CALDB environment variable is used for
    the directory containing the files, otherwise the caldb argument
    is used.
    """

    if caldb == None:
        caldb = get_caldb_dir()
    fname = caldb + "/docs/chandra/caldb_version/caldb_version.fits"
    cr = pycrates.TABLECrate(fname)

    for cname in ["CALDB_VER", "CALDB_DATE"]:
        if not cr.column_exists(cname):
            raise IOError("Unable to find the {0} column in the CALDB version file.".format(cname))

    cversion = pycrates.copy_colvals(cr, "CALDB_VER")[-1].strip()
    cdate    = pycrates.copy_colvals(cr, "CALDB_DATE")[-1].strip()

    _ciao44_crates_fix(cr)
    return (cversion, todate(cdate))

def get_caldb_installed_version(caldb=None):
    """Return the version of the installed Chandra CALDB as a string.

    If caldb is None then the CALDB environment variable is used for
    the location of the database, otherwise the caldb argument is
    used.
    """

    return get_caldb_installed(caldb)[0]

def get_caldb_installed_date(caldb=None):
    """Return the date of the installed Chandra CALDB as a time
    object.

    If caldb is None then the CALDB environment variable is used for
    the location of the database, otherwise the caldb argument is
    used.
    """

    return get_caldb_installed(caldb)[1]

def version_to_tuple(version):
    """Given a CALDB version string like '4.4.10' return
    a tuple of integers (4,4,10).
    """

    toks = version.split('.')
    try:
        out = [int(t) for t in toks]
    except ValueError:
        raise ValueError("Invalid version string '{}'".format(version))

    return tuple(out)

def check_caldb_version(version=None):
    """Checks whether the locally-installed Chandra CALDB installation
    is up-to-date. The routine requires that the computer is on-line
    and able to access the Chandra CALDB web site
    (http://cxc.harvard.edu/caldb/).

    The version argument gives the version to check against (it should be
    a string such as "4.2.2"); if None, the version of the
    local CALDB installation is used.

    If the installation is up to date then the routine returns None,
    otherwise it returns a tuple of (version checked against, latest version).
    """
    
    # Converting the dotted string form (4.2.1 or 3.2.0.1) to
    # a tuple of integers means we can just use Python's comparison
    # operator and it handles cases like
    #
    # chips-24> (4,2,1) > (3,2,0,1)
    #           True
    # chips-25> (4,2,2) > (4,2,2,2)
    #           False
    #
    # We are relying on the CALDB release numbers to be dotted integers,
    # with no alphanumerics (e.g. 4.4.1a)
    #
    rels = get_caldb_releases()
    if version == None:
        installed_ver = get_caldb_installed_version()
    else:
        installed_ver = version

        # We check against the CALDB release numbers rather than
        # the CALDB ones, since they form the "complete" set.
        # Should perhaps be case insensitive but leave that for
        # a later revision.
        #
        if version not in [v[0] for v in rels["CALDB"]]:
            raise IOError("The input CALDB version '{0}' is unknown!".format(version))

    iver = version_to_tuple(installed_ver)
    out = [v for (v,d) in rels["CIAO"] if version_to_tuple(v) > iver]
    if out == []:
        return

    out.sort(reverse=True)
    return (installed_ver, out[0])

# End
