#
#  Copyright (C) 2010, 2011, 2012, 2013  Smithsonian Astrophysical Observatory
#
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""A wrapper around the cxcdm module, partly to provide extra functionality
and in part to support upgrading scripts from CIAO 4.2 to CIAO 4.3. With
the improved funtionality of pycrates in CIAO 4.4 the need for this module
is lessened but not removed.

This is meant as an internal module and so should not be relied upon
by user code, since there is no guarantee that the API will remain stable.
"""

import os
import sys
import time

from collections import namedtuple

import numpy as np
import cxcdm as dm

from ciao_contrib.logger_wrapper import initialize_module_logger

logger = initialize_module_logger("cxcdm_wrapper")
v1 = logger.verbose3
v3 = logger.verbose3
v5 = logger.verbose5

Record = namedtuple("Record", "name value unit comment")
Column = namedtuple("Column", "name pos type dims unit comment range")

__all__ = ["open_block_from_file", "close_block",
           "get_block_info",
           "get_block_info_from_file",
           "get_info_from_file",
           "get_keyword", "get_block_name", "convert_block_number",
           "get_obsfiles",
           "open_column",
           ]

def open_block_from_file(filename):
    """Open the file, returning the block"""

    v5("Opening file {0}".format(filename))
    try:
        bl = tableOpen(filename)

    except IOError:
        v5("  -- not a table, try an image")

        try:
            bl = imageOpen(filename)

        except IOError, ie:
            v5("  -- not an image")
            raise IOError("Unable to open infile='{0}'\n  {1}".format(filename, str(ie)))

    return bl

def close_block(bl):
    """This used to work around a bug in the S-Lang module; should
    not be needed in Python but left in for old code."""
    
    #v5("Closing file - via block {0}".format(blockGetName(bl)))
    #ds = blockGetDataset(bl)
    #v5(" -> file {0}".format(datasetGetName(ds)))
    #datasetClose(ds)

    v5("Closing block '{0}'".format(blockGetName(bl)))
    blockClose(bl)

def cleanval(val):
    """If val is "empty" then return None else val."""
    if str(val).strip() == "":
        return None
    else:
        return val

def get_block_info(bl):
    """Returns a tuple with the block name and a dictionary containing
    information about the block.

    """

    blname = blockGetName(bl)
    bltype = blockGetType(bl)

    blinfo = {
        "name": blname,
        "number": blockGetNo(bl),
        }

    if bltype == dmIMAGE:
        di = imageGetDataDescriptor(bl)
        adims = getArrayDimensions(di)
        if adims == None:
            blinfo["type"] = "PRIMARY"

        else:
            blinfo["type"] = "IMAGE"
            blinfo["datatype"] = getDataType(di)
            blinfo["shape"] = adims

    elif bltype == dmTABLE:

        # assume a table with no columns is an attempt to read
        # an arbitrary ASCII file
        #
        nc = tableGetNoCols(bl)
        if nc == 0:
            blinfo["type"] = "EMPTY-TABLE"
            
        else:
            blinfo["type"] = "TABLE"
            blinfo["nrows"] = tableGetNoRows(bl)
            blinfo["columns"] = dict([(getName(c),
                                       Column(getName(c),
                                              p,
                                              getDataType(c),
                                              getArrayDimensions(c).tolist(),
                                              cleanval(getUnit(c)),
                                              cleanval(getDesc(c)),
                                              descriptorGetRange(c)
                                                       ))
                                      for (c,p) in
                                      zip(tableOpenColumnList(bl), range(1, nc+1))])

    else:
        blinfo["type"] = "UNKNOWN ({0})".format(bltype)

    # We convert from np.string_ to basic Python string since I have
    # seen some issues with the NumPy string type when sending empty
    # values to parallel processes using multiprocessing.
    #
    keyinfo = { "HDUNAME": Record("HDUNAME", blname, None, None) }
    for key in blockGetKeyList(bl):
        keyname = getName(key)
        keyval = getData(key)
        if type(keyval) == np.string_:
            keyval = str(keyval)
        keytype = getDataType(key)
        keycom = cleanval(getDesc(key))
        keyunit = cleanval(getUnit(key))
        keyinfo[keyname] = Record(keyname, keyval,
                                  comment=keycom, unit=keyunit)

    blinfo["records"] = keyinfo

    return (blname, blinfo)

def get_block_info_from_file(filename):
    """Return the tuple (blockname, blockinfo) for the
    "most-interesting" block in filename (which may include a block
    specifier).

    See get_info_from_file() for access to all blocks in a file.

    blockinfo is a dictionary with the following keys:

      name     - block name
      number   - block numner (starting from 1)
      type     - block type ("IMAGE", "TABLE" or "PRIMARY")
      records  - a dictionary of the block keywords where the values
                 are Record objects; some "structural" keywords will
                 not be included in this dictionary

    If the block is a TABLE then the following keys will also be
    present:

      nrows    - the number of rows
      columns  - a dictionary of Column objects listing the table columns;
                 the key is the column name

    If the block is an IMAGE then the following keys will also be
    present:

      datatype - a numpy data type
      shape    - a numpy array giving the image dimensions

    The Record object has attributes: name, value, unit and comment.
    The Column object has attributes: name, pos, type, dims, unit, comment, and range

    Both classes are created by the collections.namedtuple routine.

    """

    v3("Reading block information from {}".format(filename))
    bl = open_block_from_file(filename)
    try:
        out = get_block_info(bl)

    finally:
        close_block(bl)

    return out

def get_info_from_file(filename):
    """Return an array of (blockname, blockinfo) tuples for all the
    blocks in the file. See the description of get_block_info_from_file
    for a description of the blockinfo value.

    """

    ds = datasetOpen(filename)
    if ds == None:
        raise IOError("Unable to open file '{0}'\n  ".format(
                filename, getErrorMessage()))

    try:
        nblocks = datasetGetNoBlocks(ds)
        v5("File {0} has {1} blocks".format(filename, nblocks))

        out = []
        for i in range(1, nblocks + 1):
            try:
                bl = datasetMoveToBlock(ds, i)

            except IOError:
                datasetClose(ds)
                raise IOError("Unable to move to block number {0} in {1}".format(
                    i, filename))

            # Should we close the block?
            out.append(get_block_info(bl))

    finally:
        datasetClose(ds)

    return out

def get_keyword(bl, keyword):
    """Returns the value of the keyword in the block. A ValueError is
    raised if the keyword does not exist."""

    try:
        dd = keyOpen(bl, keyword)

    except ValueError:
        raise ValueError("No {0} keyword found in input file.".format(keyword))

    return getData(dd)

def get_block_name(filename, blnum):
    """Return the name of block number blnum of the
    given file, where blnum uses the CXCDM scheme
    (first block is 1)."""

    v5("Looking for name of block {0} in '{1}'".format(blnum,filename))
    ds = datasetOpen(filename);

    try:
        bl = datasetMoveToBlock(ds, blnum)

    except IOError:
        datasetClose(ds)
        raise IOError("Unable to move to block number {0} of '{1}'".format(blnum, filename))

    name = blockGetName(bl)
    datasetClose(ds)
    v5(" -> block name='{0}'".format(name))
    return "{0}[{1}]".format(filename, name)

def convert_block_number(filename, system="cxcdm", insystem="cxcdm"):
    """Given a file name that may include a CXCDM or CFITSIO style
    block number (specified by the insystem parameter), convert it to
    a new system given by the system parameter:

      none    - removes the block specifier
      name    - replace with the block name
      cxcdm   - use the CXCDM numbering scheme (starts at 1)
      cfitsio - use the CFITSIO scheme (starts at 0).

    The insystem argument can only be cxcdm or cfitsio. If system == insystem
    then the input filename is returned with no processing.
    
    We only look at the text within the last [] pair in the filename,
    and only if it is the last element of the filename.

    """

    systems = ["none", "name", "cxcdm", "cfitsio"]
    insystems = systems[2:]
    if system not in systems:
        raise ValueError("Unrecognized system={0}, should be one of:\n    {1}".format(system, " ".join(systems)))

    if insystem not in insystems:
        raise ValueError("Unrecognized insystem={0}, should be one of:\n    {1}".format(insystem, " ".join(insystems)))

    if filename == "" or filename[-1] != ']' or system == insystem:
        return filename

    lval = filename.rfind("[")
    if lval == -1:
        return filename

    head = filename[:lval]
    oval = filename[lval+1:-1]
    try:
        nval = int(oval)
    except ValueError:
        raise ValueError("Expected [<integer>] but found [{0}] in:\n{1}\n".format(oval, filename))

    # Ensure nval is in the CXCDM system.
    if insystem == "cfitsio":
        nval = nval + 1

    if system == "none":
        v5("Stripping off block number from '{0}'".format(filename))
        return head

    elif system == "cfitsio":
        v5("Converting block number to CFITSIO format in '{0}'".format(filename))
        return "{0}[{1}]".format(head, nval-1)

    elif system == "cxcdm":
        v5("Converting block number to CXCDM format in '{0}'".format(filename))
        return "{0}[{1}]".format(head, nval)

    else:
        v5("Replacing block number by name in '{0}'".format(filename))
        return get_block_name(head, nval)

def get_obsfiles(filename, filetypes, exist=True):
    """Get the full paths to the Chandra-observation filetypes
    listed in filename (which is assumed to be a Chandra event
    file but need not be as all that is looked for are the *FILE
    keywords in its header).

    filetypes is an array of names, such as ["asol", "bpix1"], for the
    Chandra observation files wanted. The names are case insensitive and
    are taken to match the *FILE keywords stored in the header.

    The return value is a dictionary with key=filetype value, value=array
    of filenames. Each filename will have been given an absolute path.
    The key matches the case of the value used in the filetypes array.

    If exist is True then each filename in the return structure will be
    checked to see if it exists and an IOError raised if it is not. When
    exist is False no check is made. Note that the check is just os.path.exists(),
    it does not try to read in and validate the file.
    
    """

    bpath = os.path.dirname(os.path.abspath(filename))
    bl = open_block_from_file(filename)
    out = {}
    try:
        
        for ftype in filetypes:
            fname = ftype.upper()
            val = get_keyword(bl, "{0}FILE".format(fname))

            # assuming that a simple split-on-comma approach will work here
            fs = [os.path.normpath(os.path.join(bpath, v)) for v in val.split(",")]
            
            if exist:
                for f in fs:
                    if not os.path.exists(f):
                        raise IOError("Unable to find {0}FILE from {1}: {2}".format(fname, filename, f))

            out[ftype] = fs
            
    finally:
        blockClose(bl)

    return out

# In CIAO 4.4 cxcdm does not raise a useful error if colname
# is invalid.
#
def open_column(bl, colname, filename=None):
    """Open the given column. This is only needed because
    it adds nicer error handling to tableOpenColumn. If
    filename is not None then it is used in debug/error messages.
    """

    try:
        if filename == None:
            msg = "Opening column {}".format(colname)
        else:
            msg = "Opening column {} in {}".format(colname, filename)

        v3(msg)
        return tableOpenColumn(bl, colname)
    
    except RuntimeError:
        if filename == None:
            msg = "There is no column called {}".format(colname)
        else:
            msg = "There is no column called {} in {}".format(colname, filename)

        raise IOError(msg)
    
### clear_acis_status_bits: START
#    
# The following routines are used by clear_acis_status_bits
# and chandra_repro. They are not exported by default.
#

# Clear bits 1-5, 
#            14-16
#            17-20, 23
#
# The bits are numbered from the right, starting from 0,
# so the above is - where x means pass through -
#
#    xxxxxxxx 0xx00000 00xxxxxx xx00000x
#
_cbits = np.uint8([0b11111111, 0b01100000, 0b00111111, 0b11000001])
_cbits_str = np.binary_repr(_cbits[3] | _cbits[2] << 8 | _cbits[1] << 16 | _cbits[0] << 24)

def _clear_status_column(dd, nrows):
   """dd is a data descriptor for a 32-bit STATUS column.
   We clear the bits 1-5, 14-20, and 23 (set to 0).

   We start by doing a per-row loop to see how slow that
   runs (the idea is to avoid reading in too much of the file
   in at once, if that is possible with this interface).
   
   """

   # Would like to chunk up into the optimal buffer size
   # but I do not know how to get that information from
   # the cxcdm API, or whether it actually makes a difference.
   #
   # However, in CIAO 4.4 getData leaks memory per call so
   # we read in all the data at once to reduce this memory
   # loss.
   #
   #rstep = 1
   rstep = nrows
   for i in range(1, nrows+1, rstep):
      idata = getData(dd, i, rstep)
      ndata = np.bitwise_and(_cbits, idata)
      setData(dd, ndata, i)

def clear_acis_status_bits(fname, toolname):
   """Clear bits 1-5, 14-20, and 23 of the STATUS
   column of fname. The edits are made in place.

   toolname is the name of the tool written to the HISTORY header,
   and the CLSTBITS keyword will be added/modified to indicate
   the bit-mask used.

   """

   timeVal = time.strftime("%Y-%m-%dT%H:%M:%S")
   bl = tableOpen(fname, update=True)

   try:
      col = tableOpenColumn(bl, "STATUS")
   except RuntimeError:
      tableClose(bl)
      raise IOError("No STATUS column found in {0}".format(fname))

   nrows = tableGetNoRows(bl)
   _clear_status_column(col, nrows)

   # Add a note to the history records; we do not try to replicate
   # the CXCDM history block as we have no parameter file.
   #
   keyWrite(bl, "CLSTBITS", "{0}".format(_cbits_str),
            desc="0 means clear STATUS bit")
   blockWriteComment(bl, "HISTORY", "TOOL: {0} at {1}".format(toolname, timeVal))
   blockWriteComment(bl, "HISTORY", "  infile={0}".format(fname))
   
   tableClose(bl)
   
### clear_acis_status_bits: END

funcnames = [
    #"getErrorMessage",

    "blockGetType",
    "blockGetDataset",
    "blockGetName",
    "blockGetNoSubspaceCpts",
    "blockSetSubspaceCpt",
    #"blockGetSubspace",
    "blockGetKeyList",
    "blockGetNo",
    #"blockGetNoKeys",
    #"blockMoveToKeyNo",
    #"blockAdvanceKeys",
    "blockGetKey",
    "blockMoveToKey",
    "blockClose",
    "blockWriteComment",

    "getArrayDim",
    "getArrayDimensions",
    "getArraySize",
    "getName",
    "getElementDim",
    "getCpt",
    "getData",
    "getDesc",
    "getUnit",
    "getDataType",
    #"getDataTypeName",

    "descriptorGetType",
    "descriptorGetRange",

    "arrayGetNoAxisGroups",
    "arrayGetAxisGroup",

    "keyOpen",
    "keyWrite",

    "tableOpen",
    "tableClose",
    "tableOpenColumn",
    "tableOpenColumnList",
    "tableGetNoRows",
    "tableGetNoCols",

    "imageOpen",
    "imageClose",
    "imageGetDataDescriptor",

    "datasetOpen",
    "datasetMoveToBlock",
    "datasetClose",
    "datasetGetName",
    "datasetGetNoBlocks",
    "datasetGetCurrentBlockNo",
    "datasetNextBlock",

    "setData",

    "subspaceColGet",
    ]

constants = [
    #"dmBOOL",
    #"dmDOUBLE",
    #"dmLONG",
    #"dmTEXT",

    "dmTABLE",
    "dmIMAGE",
    #"dmIMAGEDATA",
    ]

for funcname in funcnames:
    setattr(sys.modules[__name__], funcname, getattr(dm, "dm"+funcname[0].upper()+funcname[1:]))

for const in constants:
    setattr(sys.modules[__name__], const, getattr(dm, const))

__all__ = tuple(__all__ + funcnames + constants)
