#
#  Copyright (C) 2012  Smithsonian Astrophysical Observatory
#
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Image-specific Crates routines.

At present there is only one routine - imextent.

"""

from pytransform import LINEAR2DTransform

__all__ = ('imextent', )

def imextent(img, xmin, xmax, ymin, ymax, limits='center'):
    """Create a 2D linear transform object that converts the
    image axes to cover the range xmin to xmax and
    ymin to ymax.

    The main aim is to allow you to mimic the extent
    argument of matplotlib's imshow command - e.g.

    >>> yi, xi = np.mgrid[10:20:20j, 40:60:40j]
    >>> zi = 100.0 / np.sqrt((xi-45.62)**2 + (yi- 14.7)**2)
    >>> add_image(np.log10(zi), imextent(zi, 40, 60, 10, 20))
    >>> set_image(['depth', 50])
    >>> set_plot_aspect_ratio('fit')

    although the transform object returned by imextent also
    allows you to convert between pixel/logical and physical
    coordinates.
    
    For an image whith a shape of (ny, nx):

    If limits='center' then xmin,ymin refers to the center
    of the bottom-left pixel (1,1 in FITS/logical coordinates)
    and xmax,ymax refers to the center of the top-right
    pixel (nx,ny in FITS/logical coordinates).

    If limits='edge' then xmin,ymin refers to the bottom-left
    corner of the bottom-left pixel (0.5,0.5 in FITS/logical
    coordinates) and xmax,ymax refers to the top-right corner
    of the top-right pixel (nx+0.5,ny+0.5 in FITS/logical
    coordinates).

    As an example, we create an image with 4 pixels on the
    X axis and 3 on the Y axis, then display it using the
    axis ranges x=10 to 13 and y=-10 to -6 using both
    limits=edge and center:

    >>> img = np.arange(0, 12).reshape(3,4)
    
    >>> # limits=center
    >>> add_window()
    >>> add_image(img, imextent(img, 10, 13, -10, -6))
    >>> set_image(['depth', 50])
    >>> set_plot_aspect_ratio('fit')
    >>> get_plot_xrange()
        [9.5, 13.5]
    >>> get_plot_yrange()
        [-11.0, -5.0]

    >>> # limits=edge
    >>> add_window()
    >>> add_image(img, imextent(img, 10, 13, -10, -6, limits='edge'))
    >>> set_image(['depth', 50])
    >>> set_plot_aspect_ratio('fit')
    >>> get_plot_xrange()
        [10.0, 13.0]
    >>> get_plot_yrange()
        [-10.0, -6.0]

    In the limits=center case, the pixels are centered at x=10,11,12
    and y=-10,-8,-6 and so have widths of 1 and heights of 2. The
    plot limits therefere cover x=10-0.5 to 12+0.5 and
    y=-10-1 to -6+1.

    In the limits=edge case, the x axis covers x=10 to 13 and so
    the pixels have widths of 3.0/4 and heights of 4.0/3. The
    x and y centers of the pixels are therefore
      x=10 + 3.0/8, 10 + 9.0/8, 10 + 15.0/8, 10 + 21./8
      y=-10 + 2.0/3, -10 + 6.0/3, -10 + 10.0/3

    """

    try:
        (ny, nx) = img.shape

    except AttributeError:
        raise ValueError("First argument has no shape attribute.")

    dx = (xmax - xmin) * 1.0
    dy = (ymax - ymin) * 1.0

    if limits == 'center':
        dx /= (nx - 1.0)
        dy /= (ny - 1.0)
        x0 = xmin - dx
        y0 = ymin - dy

    elif limits == 'edge':
        dx /= nx
        dy /= ny
        x0 = xmin - dx / 2.0
        y0 = ymin - dy / 2.0

    else:
        raise ValueError("limits must be 'center' or 'edge', not '{}'".format(limits))
    
    tr = LINEAR2DTransform()

    tr.get_parameter('ROTATION').set_value(0.0)
    tr.get_parameter('SCALE').set_value([dx, dy])
    tr.get_parameter('OFFSET').set_value([x0, y0])

    return tr

