# 
#  Copyright (C) 2008, 2009, 2010, 2011  Smithsonian Astrophysical Observatory
#
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Script:
  lightcurves.py

Aim:
  Provide simple light-curve cleaning routines aimed at removing
  flares from Chandra light curves, in particular background light curves.

  This is a combination of code based on the lc_clean routine from
  Maxim Markevitch (see http://hea-www.harvard.edu/~maxim/axaf/acisbg/)
  and a simple sigma-clipped algorithm

Threads:
  http://cxc.harvard.edu/ciao/threads/acisbackground/
  http://cxc.harvard.edu/ciao/threads/filter_ltcrv/

"""

# With the CIAO 4.4 release we now rely on the presence of the
# CIAO contributed scripts package
#
import os
import tempfile

import numpy as np

import pycrates as pcr
import pychips.all as pch

from ciao_contrib.runtool import dmcopy, dmgti

__all__ = ("lc_sigma_clip", "lc_clean")

__revision = "December 2011"

# CIAO 4.4 Crates workaround to ensure data is closed/memory reclaimed.
def _ciao44_crates_fix(cr):
    cr.get_dataset().snip()

def _get_default_depth():
    "Returns the default depth value from ChIPS as an integer"
    return int(pch.get_preference("default.depth"))

def _protect_chips_label(label):
    """Return a string that can be displayed by ChIPS.
    At present all that is done is to protect underscore
    characters (i.e. so that they do not signify a sub-script)."""

    if label == None:
        rval = label
    else:
        rval = label.replace("_","\\_")

    return rval

def _add_plot(pname):
    """Add a plot with an id of pname if it doesn't exist"""

    try:
        pch.add_plot(["id", pname])
    except RuntimeError as (e):
        emsg = "chips ERROR: The id {0} is currently being used by another plot\n".format(pname)
        if str(e) == emsg:
            pass
        else:
            raise

def _get_key_as_label(cr, keyname):
    """Get the value of the key from the Crate and turn it into a
    ChIPS-displayable label. Returns None if the key does not exist."""

    return _protect_chips_label(cr.get_key_value(keyname))

def _write_gti_text(outfile, tstart, tend):
    """Create a GTI file, in TEXT/DTF format, using the
    tstart/tend values, and then does a dmcopy on it
    to convert it to FITS format (to avoid some issues that
    were found in trying to use a TEXT/DTF format GTI file
    to filter an event file; these may have been fixed by
    now).
    """

    # Could use a crate to avoid having to create the output
    # manually, but it is not clear to me how this can be done
    # (since crates does not, as of CIAO 4.4, propogate
    # subspace information), so do it manually.
    #
    (ofd, tname) = tempfile.mkstemp(text=True)
    os.write(ofd, "#TEXT/DTF\n")

    def card(name, value, comment=None, unit=None):
        "Add the record card to the file"
        os.write(ofd, "{:<8s}= ".format(name))
        
        # following checks not ideal but sufficient for now
        if isinstance(value, basestring):
            sfmt = '"{}"'
            svalue = value
        elif isinstance(value, bool):
            sfmt = "{:20s}"
            if value:
                svalue = "T"
            else:
                svalue = "F"
        elif isinstance(value, (int, long)):
            sfmt = "{:20d}"
            svalue = value
        else:
            # assume a float
            sfmt = "{:20.13e}"
            svalue = value

        os.write(ofd, sfmt.format(svalue))
        if comment == None and unit == None:
            os.write(ofd, "\n")
            return

        os.write(ofd, " /")
        if unit != None:
            os.write(ofd, " [{}]".format(unit))
        if comment != None:
            os.write(ofd, " " + comment)
        os.write(ofd, "\n")

    card("XTENSION", "TABLE")
    card("HDUNAME", "FILTER")
    card("EXTNAME", "FILTER")
    card("TFIELDS", 1)
    card("HDUCLASS", "ASC")
    card("HDUCLAS1", "FILTER")
    card("ORIGIN", "ASC", comment="Source of FITS file")
    card("CREATOR", "lightcurves - {}".format(__revision), comment="tool that created this output")
    card("CONTENT", "GTI", comment="Data product identification")
    card("TTYPE1", "TIME", comment="Time column")
    card("TFORM1", "1D", comment="data format of field.")
    card("TUNIT1", "s", comment="physical unit of field.")
    card("DSTYP1", "TIME", comment="DM Keyword: Descriptor name.")
    card("DSVAL1", "TABLE", unit="s")
    card("DSFORM1", "D", comment="DM Keyword: Descriptor datatype.")
    card("DSUNIT1", "s", comment="DM Keyword: Descriptor unit.")
    card("DSREF1", ":GTI")
    os.write(ofd, "END\n\n\n")

    card("XTENSION", "TABLE")
    card("HDUNAME", "GTI")
    card("EXTNAME", "GTI")
    card("TFIELDS", 2)
    card("TTYPE1", "START")
    card("TFORM1", "1D", comment="data format of field.")
    card("TUNIT1", "s", comment="physical unit of field.")
    card("TTYPE2", "STOP")
    card("TFORM2", "1D", comment="data format of field.")
    card("TUNIT2", "s", comment="physical unit of field.")
    card("CONTENT", "GTI", comment="Data product identification")
    card("HDUCLASS", "OGIP")
    card("HDUCLAS1", "GTI")
    card("HDUCLAS2", "STANDARD")
    card("ORIGIN", "ASC", comment="Source of FITS file")
    card("CREATOR", "lightcurves - {}".format(__revision), comment="tool that created this output")
    card("DSTYP1", "TIME", comment="DM Keyword: Descriptor name.")
    card("DSVAL1", "TABLE", unit="s")
    card("DSFORM1", "D", comment="DM Keyword: Descriptor datatype.")
    card("DSUNIT1", "s", comment="DM Keyword: Descriptor unit.")
    card("DSREF1", ":GTI")
    os.write(ofd, "END\n\n")

    for (tlo, thi) in zip(tstart, tend):
        os.write(ofd, "{:19.13e} {:19.13e}\n".format(tlo, thi))

    os.fsync(ofd)
    os.close(ofd)

    # Convert to FITS format and clean up, note that mkstemp makes no guarantees
    # that the filename is "clean", ie needs no protection.
    #
    dmcopy.punlearn()
    dmcopy(tname, outfile, clobber=True)
    os.unlink(tname)

# We stick pretty-much everything in a class which is a
# fairly-poor design
#
class LightCurve:
    """Store data from a lightcurve and provide
    methods to manipulate and display the data"""

    def __init__(self, filename, verbose=1):
        self.filename = filename
        self.verbose = verbose == 1
        self.__read_data()

        # The storage is rather redundant here (e.g. filter and clean_gti
        # are the same) but was originally written to support easy comparison
        # to the S-Lang version. Since we now the S-Lang version call the
        # Python code, this could be cleaned up.
        #
        self.clean_filter = None
        self.clean_gti = None
        self.clean_bti = None
        self.clean_min_rate = None
        self.clean_max_rate = None
        self.clean_mean_rate = None

        self.method = None
        self.userlimit = None

    def __read_data(self):
        """
        Reads in the data from the filename and performs simple validation.

        We should use TIMEUNIT and TIMEPIXR to handle non-standard cases,
        but ignore for now.
        """
        
        cr = pcr.read_file(self.filename)
    
        if cr.column_exists("count_rate"):
            self.ratename = "count_rate"
        elif cr.column_exists("rate"):
            self.ratename = "rate"
        else:
            raise IOError("No count_rate or rate column in file '{}'".format(self.filename))

        self.time = cr.get_column("time").values.copy()
        if self.time.size < 1:
            raise IOError("No data read in from the lightcurve '{}'".format(self.filename))
        elif self.time.size < 2:
            raise IOError("Only 1 time bin found in the lightcurve '{}'".format(self.filename))

        if self.verbose:
            print("Total number of bins in lightcurve   = {:d}".format(self.time.size))

        self.rate = cr.get_column(self.ratename).values.copy()

        if cr.column_exists("exposure"):
            self.exposure = cr.get_column("exposure").values.copy()
            self.bin_width = self.exposure.max()

            # We do not make use of this filter, so commenting out for now
            #filter = self.exposure <= 0.0

            if self.verbose:
                nsmall = sum( (self.exposure < self.bin_width) & (self.exposure > 0.0))
                n0 = sum(self.exposure <= 0.0)
                print("Max length of one bin                = {:g} s".format(self.bin_width))
                print("Num. bins with a smaller exp. time   = {:d}".format(nsmall))
                print("Num. bins with exp. time = 0         = {:d}".format(n0))

        else:
            self.exposure = None
            self.bin_width = None

        if cr.column_exists("time_min") and cr.column_exists("time_max"):
            self.time_min = cr.get_column("time_min").values.copy()
            self.time_max = cr.get_column("time_max").values.copy()
            self.time_offset = self.time_min[0]
        else:
            self.time_min = None
            self.time_max = None
            self.time_offset = self.time[0]

        self.labels = {}
        self.add_label(cr, "OBJECT")
        self.add_label(cr, "OBS_ID")
        self.add_label(cr, "EXPOSURE", protect=False)
        self.add_label(cr, "DTCOR", protect=False)
        self.add_label(cr, "ONTIME", protect=False)
        self.add_label(cr, "TIMEDEL", protect=False)

        self.filter = self.rate > 0.0
        if any(self.filter) == False:
            raise IOError("No rows with a count rate > 0 ({})".format(self.filename))

        # Is there a better way to do this?
        self.gti = self.filter
        self.bti = np.logical_not(self.filter)

        if self.verbose and any(self.bti):
            print("Number of bins with a rate of 0 ct/s = {}\n".format(self.bti.sum()))

        self.mean_rate_original = self.rate.mean()
        self.mean_rate_filtered = self.rate[self.gti].mean()

        _ciao44_crates_fix(cr)

    def add_label(self, crate, name, protect=True):
        """If there is a keyword of the given name in the crate, add
        it to the store. If protect is True then ensure the value is
        'protected' for display by ChIPS."""

        if pcr.key_exists(crate, name):
            val = pcr.get_keyval(crate, name)
            if protect: val = _protect_chips_label(val)
            self.labels[name] = val

    def time_to_offset(self, t):
        "Convert the time(s) in t, assumed to be in seconds, to an offset in kiloseconds"
        return (t - self.time_offset) / 1.0e3

    def calculate_valid_time_bins(self, minlength=1):
        """Returns (tstart, tend, exposure, flag) for those time bins that pass
        the filter (so calculate_filter() must have already been called).
        flag is True if all intervals pass the minlength filter (i.e. if using
        minlength has no affect here). exposure is the exposure length of
        each interval (if the input light curve had an EXPOSURE column), otherwise
        None.

        The minlength parameter gives the minimum number of consecutive
        bins that must pass the filter to still remain valid (i.e. if
        changed from 1 it acts as an additional filter).
        """

        if self.clean_min_rate == None:
            raise StandardError("calculate_filter() must be run before calculate_valid_time_bins()")

        # Loop through the time bins and ensure that each set of
        # selected times contains >= minlength bins
        #
        if self.time_min == None:
            xlo = self.time
        else:
            xlo = self.time_min
        if self.time_max == None:
            xhi = self.time
        else:
            xhi = self.time_max

        # A simplified version of run-length encoding to calculate
        # those bins that pass both the count rate limit and the
        # minlength test.
        #
        # We add 0's onto the start and end of the filter to simplify
        # handling the start and end bins.
        #
        filter = np.concatenate(([0], self.clean_filter * 1, [0]))
        x = np.diff(filter)
        pstart, = np.where(x == 1)
        pend,   = np.where(x == -1)

        if pstart.size != pend.size:
            raise StandardError("Internal error: pstart/end lengths are {} and {}!".format(pstart.size, pend.size))

        # Filter out those intervals that do not contain minlength
        # elements. 
        #
        pnum = pend - pstart
        lcheck = pnum >= minlength
        if not np.any(lcheck):
            raise StandardError("Error: there are no periods which pass the minlength ({}) check!".format(minlength))

        i, = np.where(lcheck)
        pend -= 1
        xlo = xlo[pstart[i]]
        xhi = xhi[pend[i]]

        if self.exposure == None:
            exposure = None
        else:
            exposure = np.asarray([self.exposure[pstart[j]:pend[j]+1].sum() for j in i])

        return (xlo, xhi, exposure, np.all(lcheck))

    # Should be able to clean up this mess somewhat
    #
    def report_userlimit_using_times(self, tlo=None, thi=None, exposure=None, minlength=1):
        """Display, to the screen, the valid times as time filters along with
        the corresponding exposure times. At the end indicate the amount of
        valid exposure time that has been retained, although this is only
        approximate as it depends on the GTI intervals in the input file
        as well."""

        if self.verbose == False:
            return

        if (tlo == None) != (thi == None):
            raise StandardError("tlo and thi must either both be None or both be arrays")

        if tlo == None:
            (tlo, thi, exposure, tflag) = self.calculate_valid_time_bins(minlength)

        ctr = np.arange(tlo.size) + 1
        if exposure == None:

            exposure = (thi - tlo) / 1.0e3

            # There is no guarantee that the exposure time of this bin is the same as (thi-tlo),
            # but we include the DTCOR factor just in case
            #
            if self.labels.has_key("DTCOR"):
                exposure *= self.labels["DTCOR"]
        else:

            # Assume that the exposure values include any DTCOR factor, and are in seconds
            exposure /= 1.0e3

        # ustr    = ["  ((time >= {:f}) && (time < {:f})) ; {:.2f} ksec, bin {}".format(s,e,d,c) for c,s,e,d in zip(ctr,tlo,thi,exposure)]
        ustr    = ["  ((time >= {}) && (time < {})) ; {:.2f} ksec, bin {}".format(repr(s),repr(e),d,c) for c,s,e,d in zip(ctr,tlo,thi,exposure)]

        n = len(tlo)
        if n == 1:
            print(ustr[0])
        else:
            print("\n".join(ustr))
        print("")
        if self.labels.has_key("EXPOSURE"):
            print("  Exposure time of lightcurve = {:.2f} ks".format(self.labels["EXPOSURE"]/1000.0))
        print("  Filtered exposure time      = {:.2f} ks".format(exposure.sum()))
        if self.labels.has_key("DTCOR"):
            print("  DTCOR value                 = {:g}".format(self.labels["DTCOR"]))

    def create_userlimit_using_rates(self, minlim, maxlim):
        """Sets the userlimit field for the rate column to
        lie between minlim and maxlim."""

        self.userlimit = "({}>{} && {}<{})".format(self.ratename, repr(minlim), self.ratename, repr(maxlim))
        if self.verbose:
            print("GTI limits calculated using a count-rate filter:\n  {}\n".format(self.userlimit))
            print("The corresponding times are:")
            self.report_userlimit_using_times(minlength=1)

    # As moving towards writing out the GTI file manually for this case the
    # semantics of this routine no longer matches the name, in that we end
    # up storing the lo/hi values for use by _write_gti_text
    #
    def create_userlimit_using_times(self, lo, hi, exposure=None, minlength=1):
        "Sets the userlimit field for time values between lo and hi."

        # Hmm, should we just go for full accuracy? Not really worth it for most cases
        #ulimits = ["((time>={})&&(time<{}))".format(repr(a), repr(b)) for a,b in zip(lo,hi)]
        ulimits = ["((time>={:f})&&(time<{:f}))".format(a,b) for a,b in zip(lo,hi)]
        n = len(lo)
        if n == 1:
            self.userlimit = ulimits[0]
        else:
            self.userlimit = "({})".format("||".join(ulimits))

        # HACK: save the lo/hi/exposure values for use by _write_gti_text
        self.userlimit_bins = (lo, hi, exposure)

        if self.verbose:
            print("GTI limits calculated using a time filter:")
            self.report_userlimit_using_times(tlo=lo, thi=hi, exposure=exposure, minlength=minlength)

    def _add_plot_labels(self):
        """Add filename/object/obsid labels to the current plot,
        and make sure the borders are not displayed for this plot
        Ensure these labels are at a lower depth than the main
        plot items.

        Prior to CIAO 4.4 these labels were added to a separate plot,
        with an id of 'labels', but this has been removed.
        """

        label_depth = _get_default_depth() - 10
        if label_depth < 0: label_depth = 0

        namelabel = _protect_chips_label(self.filename.split("/")[-1])

        # pch.add_plot(["id", "labels", "style", "open"])
        pch.add_label(0.05, 0.05, namelabel, \
                          ["id","filename", "halign", 0, "coordsys", pch.FRAME_NORM, "depth", label_depth])
        pch.add_label(0.95, 0.05, _protect_chips_label(self.method), \
                          ["id","method", "halign", 1, "coordsys", pch.FRAME_NORM, "depth", label_depth])
        if self.labels.has_key("OBJECT"):
            pch.add_label(0.05, 0.95, self.labels["OBJECT"], \
                              ["id","object", "halign", 0, "coordsys", pch.FRAME_NORM, "depth", label_depth])
        if self.labels.has_key("OBS_ID"):
            pch.add_label(0.95, 0.95, "obsid=" + self.labels["OBS_ID"], \
                              ["id","obsid", "halign", 1, "coordsys", pch.FRAME_NORM, "depth", label_depth])

    def plot(self, rateaxis="y", gcol="green", erase=True):
        """Plot up the light curves.

        The rateaxis argument determines whether the rate is drawn on
        the Y axis or X axis ("y" or "x" respectively). The gcol
        argument is the color to use to draw the "good" points.

        If erase is True then erase will be called in the current
        window before creating the plots, otherwise the current
        frame will be used (only relevant if there is a current
        frame). Note that we still add plot label/decorations
        when erase is False, since it is only intended to allow
        a plot to be created in an existing, but empty, Frame 
        rather than overplotted on a previous lc plot.
        """

        if self.clean_gti == None:
            raise StandardError("calculate_filter() must be run before plot()")
        gti = self.clean_gti
        bti = self.clean_bti

        # What data are we plotting?
        #
        t = self.time_to_offset(self.time)
        x1g = t[gti]
        y1g = self.rate[gti]

        if any(bti):
            x1b = t[bti]
            y1b = self.rate[bti]
        else:
            x1b, y1b = None, None

        # We need to calculate the binning for the histograms, so we
        # pick a bin size that is the minimum of 0.01 and (rmax-rmin)/5.0
        # where rmax/min are the max/min count rates of the filtered data.
        # As we have now had reports of this causing a problem - admittedly
        # only for problematic data sets - then we now include a rather
        # arbitrary limit on the number of bins.
        #
        rmin = self.clean_min_rate
        rmax = self.clean_max_rate
        rwidth = min(0.01, (rmax-rmin)/5.0)

        # Not bothered if number of bins is off by 1 here
        dmin = self.rate.min()
        dmax = self.rate.max()
        nbins = (dmax-dmin) / rwidth
        if nbins > 1000:
            print("Warning: Default bin width of {:g} count/s is too small as it produces".format(rwidth))
            print("         {} bins.".format(nbins))
            rwidth = (dmax-dmin) / 1000.0
            print("         Replacing with a width of {:g} count/s".format(rwidth))
            print("         This may indicate that the lightcurve contains strong flares that")
            print("         require manual filtering.\n")

        ledges = np.arange(dmin, dmax, rwidth)
        ledges = np.append(ledges, ledges[-1]+rwidth)

        (y2a,dummy) = np.histogram(self.rate, bins=ledges)
        xlo2a = ledges[:-1]
        xhi2a = ledges[1:]

        (y2g,dummy) = np.histogram(y1g, bins=ledges)
        xlo2g = xlo2a
        xhi2g = xhi2a

        # Filter out zero-rate bins as they are a visual distraction
        #
        i, = np.where(y2a > 0)
        xlo2a = xlo2a[i]
        xhi2a = xhi2a[i]
        y2a = y2a[i]

        i, = np.where(y2g > 0)
        xlo2g = xlo2g[i]
        xhi2g = xhi2g[i]
        y2g = y2g[i]

        # Set up the plot labels
        #
        title = "mean rate={:g} s^{{-1}}".format(self.clean_mean_rate)
        ratelabel = "Count Rate (s^{-1})"

        # Create the plot within an undo buffer so that
        #  a) a single undo call will remove the whole plot
        #  b) if there is an error during the plot then
        #     it is easy to get back to the original ChIPS state
        #
        pch.open_undo_buffer()
        try:
            if pch.info() == None:
                pch.add_frame(["id", self.method, "window.id", "lc_plot"])
            elif erase:
                pch.erase()
                pch.add_frame(["id", self.method])

            _add_plot("rate")
            _add_plot("hist")

            if rateaxis == "y":
                spacing = 0.12
            else:
                spacing = 0.0
            pch.split(2, 1, spacing)
            pch.adjust_grid_yrelsize(1, 1.2)

            # Prior to CIAO 4.4 the labels were added to a separate plot; now
            # they are added to the first plot.
            pch.current_plot("rate")
            self._add_plot_labels()

            # plot 1: lightcurve
            #
            # The "bad" set of points are added first, if they
            # exist, so that they are drawn behind the "good" points
            #
            pch.current_plot("rate")
            pch.set_plot_title(title)

            copts = {"line.style": "noline", "symbol.fill": True}
            
            if any(bti):
                if rateaxis == "y":
                    x = x1b
                    y = y1b
                else:
                    x = y1b
                    y = x1b

                copts["id"] = "bad"
                copts["symbol.style"] = "square"
                copts["symbol.size"] = 2
                pch.add_curve(x, y, copts)

            if rateaxis == "y":
                x = x1g
                y = y1g
            else:
                x = y1g
                y = x1g

            copts["id"] = "good"
            copts["symbol.style"] = "circle"
            copts["symbol.size"] = 4
            copts["symbol.color"] = gcol
            pch.add_curve(x, y, copts)

            if rateaxis == "y":
                pch.add_hline(self.clean_mean_rate, ["id", "hline", "style", "dot"])

                pch.set_plot_xlabel(r"\Delta Time (ks)")
                pch.set_plot_ylabel(ratelabel)
            else:
                pch.add_vline(self.clean_mean_rate, ["id", "hline", "style", "dot"])

                pch.set_xaxis(["ticklabel.visible", False])
                pch.set_plot_ylabel(r"\Delta Time (ks)")

            # plot 2: histogram
            #
            pch.current_plot("hist")
            if any(bti):
                pch.add_histogram(xlo2a, xhi2a, y2a, ["id", "everything"])

            pch.add_histogram(xlo2g, xhi2g, y2g, ["id", "good", "line.color", gcol, "line.thickness", 3])

            pch.add_vline(self.clean_mean_rate, ["id", "vline", "style", "dot"])

            pch.limits(pch.Y_AXIS, 0, pch.AUTO)
            pch.set_plot_xlabel(ratelabel)
            pch.set_plot_ylabel("Number")
            pch.limits(pch.X_AXIS, pch.AUTO, pch.AUTO)

        except:
            pch.discard_undo_buffer()
            raise

        pch.close_undo_buffer()

    def plot_gti_limits(self, gtiname, rateaxis="y", pattern="crisscross", pcol="red"):
        "Plot up the GTI limits from the GTI block of gtiname on the rate plot"

        if pattern == "none":
            return

        cr = pcr.read_file("{}[gti]".format(gtiname))

        pch.open_undo_buffer()
        try:
            
            pch.current_plot("rate")

            if rateaxis == "y":
                xr = pch.get_plot_xrange()
                yr = pch.get_plot_yrange()
            else:
                xr = pch.get_plot_yrange()
                yr = pch.get_plot_xrange()

            yrs = [yr[0], yr[0], yr[1], yr[1]]

            xstart = np.append(self.time_to_offset(pcr.copy_colvals(cr,"start")), xr[1])
            xend   = np.append(xr[0], self.time_to_offset(pcr.copy_colvals(cr,"stop")))

            reg_depth = _get_default_depth() - 10
            if reg_depth < 0: reg_depth = 0

            ropts = ["fill.style", pattern, "fill.color", pcol, "opacity", 1.0, \
                         "edge.style", "none", "edge.color", pcol, \
                         "depth", reg_depth]

            for xs,xe in zip(xend,xstart):
                if rateaxis == "y":
                    pch.add_region([xs,xe,xe,xs], yrs, ropts)
                else:
                    pch.add_region(yrs, [xs,xe,xe,xs], ropts)

            pch.current_plot("hist")

        except:
            pch.discard_undo_buffer()
            raise

        pch.close_undo_buffer()

        _ciao44_crates_fix(cr)

    def create_gti_file(self, outfile):
        """Create a GTI file called outfile based on those time periods
        from infile that match the given filter (which is a string of
        the form accepted by the userlimit parameter of dmgti
        """

        if self.userlimit == None:
            raise StandardError("calculate_gti_filter() must be run before create_gti_file()")

        if self.verbose:
            print("\nCreating GTI file")

        # Do we write out the GTI manually?
        if hasattr(self, "userlimit_bins"):
            _write_gti_text(outfile, self.userlimit_bins[0], self.userlimit_bins[1])
        else:
            dmgti.punlearn()
            dmgti(self.filename, outfile, self.userlimit, clobber=True)

        if self.verbose:
            print("Created: {}".format(outfile))

class CleanLightCurve(LightCurve):
    "Light curve filtering using the same method as the ACIS background files"

    def __init__(self, filename, verbose=1):
        LightCurve.__init__(self, filename, verbose=verbose)
        if self.exposure == None:
            raise IOError("The lightcurve '{}' does not contain an EXPOSURE column!".format(filename))
        self.method = "lc_clean"

    def check_valid(self, minfrac):
        """Ensure that the fraction of good bins is >= minfrac (excluding
        the 0-rate bins from the calculation). Throws a StandardError if
        the condition is not met."""

        nrows = self.filter.sum()
        ngood = self.clean_filter.sum()
        frac = (ngood+0.0) / nrows
        if frac < minfrac:
            raise StandardError("Fraction of bins that are good ({:g}, {} of {}) is below limit of {:g}!".format(frac, ngood, nrows, minfrac))
        
    def calculate_filter(self, mean=None, clip=3.0, sigma=None, scale=1.2):
        "Filter the light curve."

        # Do we need to calculate an initial mean level?
        #
        if mean == None:

            # The sigma is calculated assuming "poisson" statistics for the mean value,
            # so we have sqrt (mean-rate * bin-width) and then we want to convert it
            # back to a rate, to give sqrt (mean-rate / bin-width)
            #   
            omean = self.rate[self.filter].mean()
            osigma = np.sqrt(omean / self.bin_width)
            minval = omean - clip * osigma
            maxval = omean + clip * osigma

            gti, = np.where((self.rate > minval) & (self.rate < maxval))
            if gti.size == 0:
                raise StandardError("Unable to calculate an initial mean level via sigma clipping.")
            mean = self.rate[gti].mean()

            if self.verbose:
                print("Calculated an initial mean (sigma-clipped) rate of {:g} ct/s".format(mean))

        else:
            if self.verbose:
                print("Using a fixed mean rate of {:g} ct/s".format(mean))

        # Calculate the limits based on the mean value and sigma/scale clipping
        #
        if sigma == None:
            self.clean_min_rate = mean / scale
            self.clean_max_rate = mean * scale
            if self.verbose:
                print("Lightcurve limits use a scale factor of {:g} about this mean".format(scale))

        else:
            # divide by the bin width so we can assume sigma=sqrt(signal)
            #
            sigval = np.sqrt(mean / self.bin_width)
            self.clean_min_rate = mean - sigma * sigval
            self.clean_max_rate = mean + sigma * sigval
            if self.verbose:
                print("Lightcurve limits clipped using {:g} sigma's about this mean".format(sigma))

        if self.verbose:
            print("Filtering lightcurve between rates of {:g} and {:g} ct/s".format(self.clean_min_rate, self.clean_max_rate))
        
        self.clean_filter = (self.rate > self.clean_min_rate) & (self.rate < self.clean_max_rate)
        ngood = self.clean_filter.sum()
        if ngood == 0:
            raise StandardError("Error: no bins match rate={:g} to {:g} (data range is {:g} to {:g})".format(self.clean_min_rate, self.clean_max_rate,
                                                                                                              self.rate[self.filter].min(), self.rate.max()))
        if self.verbose:
            print("Number of good time bins = {}".format(ngood))

        self.clean_gti = self.clean_filter
        self.clean_bti = np.logical_not(self.clean_gti)

        self.clean_mean_rate = self.rate[self.clean_gti].mean()
        if self.verbose:
            print("Rate filter:  {} <= {} < {} ".format(repr(self.clean_min_rate), self.ratename, repr(self.clean_max_rate)))
            print("Mean level of filtered lightcurve = {} ct/s\n".format(repr(self.clean_mean_rate)))

    def calculate_gti_filter(self):
        "Calculate the GTI filter limit (the userlimit string for dmgti)"

        if self.clean_min_rate == None:
            raise StandardError("calculate_filter() must be run before calculate_gti_filter()")

        self.create_userlimit_using_rates(self.clean_min_rate, self.clean_max_rate)

class SigmaClipLightCurve(LightCurve):
    "Provide an iterative sigma-clipping filter for a lightcurve"

    def __init__(self, filename, verbose=1):
        LightCurve.__init__(self, filename, verbose=verbose)
        self.method = "lc_sigma_clip"
        
    def calculate_filter(self, sigma=3.0, minlength=3):
        "Filter the light curve."

        f = self.filter.copy()
        while True:

            # gp gives the "good points" in this iteration
            (gp,) = np.where(f)
            if gp.size == 0:
                raise StandardError("Error: no bins pass the filter")
            rate = self.rate[gp]
            mean = rate.mean()
            stdev = rate.std()

            min_rate = mean - sigma * stdev
            max_rate = mean + sigma * stdev

            asigmas = np.abs(self.rate - mean) / stdev
            (i,) = np.where((asigmas > sigma) & f)
            if i.size == 0:
                break

            f[i] = False

        # Need to improve storage
        #
        self.clean_filter = f
        self.clean_gti = f
        self.clean_bti = np.logical_not(f)
        self.clean_min_rate = min_rate
        self.clean_max_rate = max_rate
        self.clean_mean_rate = self.rate[self.clean_gti].mean()

        if self.verbose:
            print("Rate filter:  {} <= {} < {} ".format(repr(self.clean_min_rate), self.ratename, repr(self.clean_max_rate)))
            print("Mean level of filtered lightcurve = {} ct/s\n".format(repr(self.clean_mean_rate)))

    def calculate_gti_filter(self, minlength=3):
        "Caculate the GTI filter limit (the userlimit string for dmgti)"

        if self.clean_min_rate == None:
            raise StandardError("calculate_filter() must be run before calculate_gti_filter()")

        # If minlength is 1 then we can just the count-rate limits directly. Otherwise
        # we check to see whether the minlength constraint actually makes a difference
        # to the count-rate limit, only using the explicit time ranges if it does
        #
        if minlength == 1:
            self.create_userlimit_using_rates(self.clean_min_rate, self.clean_max_rate)
        else:
            (tlo, thi, tdur, tflag) = self.calculate_valid_time_bins(minlength=minlength)
            if tflag:
                self.create_userlimit_using_rates(self.clean_min_rate, self.clean_max_rate)
            else:
                self.create_userlimit_using_times(tlo, thi, exposure=tdur, minlength=minlength)

# Public access to the filtering
#

def lc_clean(filename, outfile=None, mean=None, clip=3.0, sigma=None,
             scale=1.2, minfrac=0.1,
             plot=True, rateaxis="y", pattern="solid", gcol="green",
             pcol="red", erase=True,
             verbose=1):
    """
    Calculate good times for a light curve using the same filtering as
    was used to calculate the ACIS blank-sky backgropund files.

    The filename argument must be a table containing TIME, COUNT_RATE (or
    RATE), and EXPOSURE columns.

    If outfile is set then a GTI file will be created representing these
    times.

    If mean is None then the mean count rate is calculted from data, using
    a single sigma clip (the clip pararameter gives the number of
    standard deviations to retain) of the data. The standard deviation is
    calculated as sqrt(mean rate/bin width).

    The mean value is then used to clip the light curve: if sigma is None
    then the limits are mean/scale to min*scale, otherwise they are
    mean - sigma*stdev to mean + sigma*stdev (stdev is again calculated
    from the mean value).

    The minfrac parameter gives the minimum fraction of the lightcurve
    bins that must pass the count-rate filter (after ignoring the zero
    count-rate bins).

    If plot is True then plots of the light curve and histograms of the
    values will be displayed. The rateaxis option determines whether the
    topl plot is drawn as count-rate on the y axis and time on the x axis
    (rateaxis='y') or the other way around (rateaxis='x'). The gcol
    parameter gives the color to use to draw points/histograms of the
    points that are considered good by the iterative clipping (ignoring
    the minlength filter).

    The erase setting is only important if a window exists and plot is
    True. If erase is True then any frames in the existing window will
    be removed (via the ChIPS erase call) before any plots are
    created, otherwise the plots will either be added to the existing
    frame. This is only intended to allow plots to be added to
    multiple frames in a window - for example

      lc_clean("lc.fits")
      add_frame()
      lc_clean("lc.hard.fits", erase=False)

    and not to allow multiple lightcurves to be drawn in the same
    plot.

    If outfile is given and plot is True then the time intervals excluded
    by the calculated GTI file are drawn on the plot as filled regions,
    using the pattern and color determined by the pattern and pcol options
    (setting pattern to 'none' stops this display).

    The verbose parameter can be set to 0 or 1 and determines whether
    there is any screen output.
    """

    # Check the arguments
    #
    if clip <= 0.0:
        raise ValueError("clip argument must be > 0, not {:g}".format(clip))
    if sigma == None:
        if scale < 1.0:
            raise ValueError("scale argument must be >= 1.0, not {:g}".format(scale))
    elif sigma <= 0.0:
        raise ValueError("sigma argument must be None or > 0, not {:g}".format(clip))
    if verbose != 0 and verbose != 1:
        raise ValueError("verbose argument must be 0 or 1, not {}".format(verbose))

    if rateaxis not in ["x", "y"]:
        raise ValueError("rateaxis argument must be 'x' or 'y', not '{}'".format(rateaxis))

    # Let the user know what arguments are being used
    #
    if verbose == 1:
        print("Parameters used to clean the lightcurve are:")
        print("  script version = {}".format(__revision))
        print("  mean           = {}".format(mean))
        print("  clip           = {:g}".format(clip))
        if sigma == None:
            print("  scale          = {:g}".format(scale))
        else:
            print("  sigma          = {:g}".format(sigma))
        print("  minfrac        = {:g}".format(minfrac))
        if outfile != None:
            print "  outfile        = {}".format(outfile)
        print("  plot           = {}".format(plot))
        if plot == True:
            print("  rateaxis       = {}".format(rateaxis))
            print("  color          = {}".format(gcol))
            if outfile != None:
                print("  pattern        = {}".format(pattern))
                if pattern != "none":
                    print("  pattern color  = {}".format(pcol))
        print("")

    lc = CleanLightCurve(filename, verbose=verbose)
    lc.calculate_filter(mean=mean, clip=clip, sigma=sigma, scale=scale)
    if plot:
        if outfile == None:
            patval = "none"
        else:
            patval = pattern
        lc.plot(rateaxis=rateaxis, gcol=gcol, erase=erase)

    # Check that we have enough data points
    #
    lc.check_valid(minfrac)

    # Calculate the GTI file if required. We calculate the limits
    # whether the file is to be created or not so as to get
    # screen output
    #
    lc.calculate_gti_filter()
    if outfile != None:
        lc.create_gti_file(outfile)

        if plot:
            lc.plot_gti_limits(outfile, rateaxis=rateaxis, pattern=pattern, pcol=pcol);

def lc_sigma_clip(filename, outfile=None, sigma=3.0, minlength=3, plot=True, 
                  rateaxis="y", pattern="solid", gcol="green",
                  pcol="red", erase=True,
                  verbose=1):
    """
    Calculate good times for a light curve by performing an iterative
    sigma-clip on the count rate column of filename. If outfile is set
    then a GTI file will be created representing these times.

    The sigma parameter determines which points are to be rejected at
    each iteration (reject if they are more than sigma times the
    standard deviation away from the mean).

    The minlength parameter determines the minimum number of consecutive
    time bins that must all pass the rate filter for the period to be
    considered 'good'. This parameter is not used within the
    sigma-clipping iteration. If minlength=1 then a filter on the count
    rate is determined instead.

    If plot is True then plots of the light curve and histograms of the
    values will be displayed. The rateaxis option determines whether the
    topl plot is drawn as count-rate on the y axis and time on the x axis
    (rateaxis='y') or the other way around (rateaxis='x'). The gcol
    parameter gives the color to use to draw points/histograms of the
    points that are considered good by the iterative clipping (ignoring
    the minlength filter).

    The erase setting is only important if a window exists and plot is
    True. If erase is True then any frames in the existing window will
    be removed (via the ChIPS erase call) before any plots are
    created, otherwise the plots will either be added to the existing
    frame. This is only intended to allow plots to be added to
    multiple frames in a window - for example

      lc_sigma_clip("lc.fits")
      add_frame()
      lc_sigma_clip("lc.hard.fits", erase=False)

    and not to allow multiple lightcurves to be drawn in the same
    plot.

    If outfile is given and plot is True then the time intervals excluded
    by the calculated GTI file are drawn on the plot as filled regions,
    using the pattern and color determined by the pattern and pcol options
    (setting pattern to 'none' stops this display).

    The verbose parameter can be set to 0 or 1 and determines whether
    there is any screen output.
    """

    # Check the arguments
    #
    if sigma <= 0.0:
        raise ValueError("sigma argument must be > 0, not {:g}".format(sigma))
    
    minlength = int(minlength) # deflare sends in a float rather than an int
    if minlength < 1:
        raise ValueError("minlength argument must be >= 1, not {:d}".format(minlength))
    if verbose != 0 and verbose != 1:
        raise ValueError("verbose argument must be 0 or 1, not {}".format(verbose))

    if rateaxis not in ["x", "y"]:
        raise ValueError("rateaxis argument must be 'x' or 'y', not '{}'".format(rateaxis))

    # Let the user know what arguments are being used
    #
    if verbose == 1:
        print("Parameters used to clean the lightcurve are:")
        print("  script version = {}".format(__revision))
        print("  sigma          = {:g}".format(sigma))
        print("  minlength      = {:d}".format(minlength))
        if outfile != None:
            print("  outfile        = {}".format(outfile))
        print("  plot           = {}".format(plot))
        print("  rateaxis       = {}".format(rateaxis))
        print("  color          = {}".format(gcol))
        if outfile != None:
            print("  pattern        = {}".format(pattern))
            if pattern != "none":
                print("  pattern color  = {}".format(pcol))
        print("")

    lc = SigmaClipLightCurve(filename, verbose=verbose)
    lc.calculate_filter(sigma=sigma, minlength=minlength)
    if plot:
        if outfile == None:
            patval = "none"
        else:
            patval = pattern
        lc.plot(rateaxis=rateaxis, gcol=gcol, erase=erase)

    # Calculate the GTI file if required. We calculate the limits
    # whether the file is to be created or not so as to get
    # screen output
    #
    lc.calculate_gti_filter(minlength=minlength)
    if outfile != None:
        lc.create_gti_file(outfile)

        if plot:
            lc.plot_gti_limits(outfile, rateaxis=rateaxis, pattern=pattern, pcol=pcol);


# End
