#
#  Copyright (C) 2009, 2010  Smithsonian Astrophysical Observatory
#
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Create a spectrum file for use in ChART.

This is a simple conversion from CIAO 3.4 to CIAO 4.1 of the
chart_spectrum.sl script. It takes advantage of the new interface
in Sherpa, which simplifies the data analysis, and the keyword
possibilities provided by Python.

"""

import sherpa.astro.ui as s

from sherpa.utils.err import IdentifierErr, ArgumentErr

import pychips.all as c

import numpy as np

import os

__all__ = ("get_chart_spectrum",
           "save_chart_spectrum",
           "plot_chart_spectrum")

def _get_chart_spectrum(id=None,elow=None,ehigh=None,ewidth=None,norm=None):
    """This performs the work of get_chart_spectrum() but returns
    extra information that is only likely to be useful to
    other high-level routines like plot_chart_spectrum."""

    # What source expression are we using?
    # get_model/source will throw an IdentifierErr if the expression
    # is not defined; we do not, at present catch/re-throw this
    #
    if id == None:
        id = s.get_default_id()

    mdl = s.get_source(id)

    # What energy grid to use?
    # Since we do not want to restrict users to only using PHA datasets
    # (i.e. if I just want to create something simple) then we have to
    # look for a range of errors from get_arf
    #
    if elow == None or ehigh == None or ewidth == None:
        try:
            arf = s.get_arf(id)
        except (IdentifierErr, ArgumentErr):
            # a) PHA dataset, no ARF
            # b) Assume this means the dataset is not derived from the PHA class
            arf = None
        except:
            raise

        if arf == None:
            raise TypeError("No ARF found for dataset {0} so unable to create energy grid".format(repr(id)))

        if elow == None:
            elow = arf.energ_lo[0]
        if ehigh == None:
            ehigh = arf.energ_hi[-1]
        if ewidth == None:
            # Assume constant grid spacing in the ARF
            ewidth = (arf.energ_hi[-1] - arf.energ_lo[0]) * 1.0 / np.size(arf.energ_lo)

    if elow >= ehigh:
        raise TypeError("elow is >= ehigh: elow={0}  ehigh={1}".format(elow,ehigh))
    if ewidth <= 0.0:
        raise TypeError("ewidth is <= 0.0: ewidth={0}".format(ewidth))

    # The following is wasteful if we have an ARF and the user
    # supplies no elow, ehigh, or ewidth arguments.
    #
    # Should I check that nbins is a sensible number (e.g. >= 2)?
    #
    nbins = 1 + np.rint( (ehigh - elow) / ewidth )
    erange = elow + ewidth * np.arange(nbins)
    elo = erange[:-1]
    ehi = erange[1:]

    # In CIAO 3.4 we had to go to a lot of trouble to get the data we wanted;
    # now we can just evaluate the model over the desired energy ranges
    flux = mdl(elo,ehi)
    emid = 0.5 * (ehi + elo)

    # do we want to renormalize?
    if norm != None:
        flux *= norm
    return { "x": emid, "y": flux, "id": id, "model": mdl.name }

def get_chart_spectrum(id=None,elow=None,ehigh=None,ewidth=None,norm=None):
    """Returns the tuple (energy_bins,flux_bins) which represent
    the spectrum of the source in a format suitable for the Chandra
    Rat Tracer (ChaRT), namely energy_bins is in keV and flux_bins
    is in photons/cm^2/s.

    If id is None then the default dataset is used (get_default_id).
    
    The elow, ehigh, and ewidth arguments represent the left edge of the
    lowest-energy bin, right edge of the highest-energy bin, and
    bin width in keV. If any are not given then the energy grid defined
    by the ARF for the source is used (if there is no ARF then an
    error is thrown).

    The norm parameter, if set, is used to multiply the flux values
    by, so norm=10.0 creates a model for a source ten times brighter
    than that of the model. This is useful if you want to create
    a high signal-to-noise PSF, since the exposure time for ChaRT
    runs has a maximum value.

    Unlike the CIAO 3.4 version we do not re-bin the data from Sherpa
    onto the grid since we can now evaluate the models directly.
    This does mean that it is possible that the start or end bins may
    extend outside the desired range, depending on the binning used by the
    ARF or the value of the ewidth parameter. For example if

      elow=2
      ehigh=2.5
      ewidth=0.2

    then we will use bins 2-2.2, 2.2-2.4, 2.4-2.6 to evaluate the source
    spectrum. The error introduced by this is assumed to be significantly
    smaller than other errors in the calculation (such as how well the
    source spectrum is actually known).

    We *assume* that the source model is made up of models that,
    when evaluated, accept keV values for the energy grid and - if
    source expressions - return values of photon/cm**2/s per bin.
    This is true of the XSpec models and 1D Sherpa models.

    TO DO: restrict evaluation to 0.2 to 10 keV range (if necessary).
    """

    vals = _get_chart_spectrum(id,elow,ehigh,ewidth,norm)
    return (vals["x"], vals["y"])

# It would have been nice to use crates_contrib.utils.write_columns
# to create the output file but there is no way to remove the header
# line giving the flavor of ASCII (e.g. '#TEXT/RAW'), and we need
# no header, so we roll our own output here (which isn't hard).
#
def save_chart_spectrum(outfile=None,verbose=True,**kwargs):
    """Creates (overwrites if it exists) the ASCII file outfile which can
    be used as the spectrum input to ChaRT.

    If outfile is None then os.sys.stdout is used for output.
    If verbose is True then a message is printed out to the screen
    indicating that a file has been created.

    The other arguments are passed to get_chart_spectrum(); see that routine
    for more information.
    """

    # The output has to be ASCII format, two columns, with no blank or
    # comment lines, so we can not use Crates for this but have to
    # write out the values manually
    #
    (emid,flux) = get_chart_spectrum(**kwargs)
    if outfile == None:
        fh = os.sys.stdout
    else:
        fh = open(outfile,"w")
    for l in map(lambda e,f: "{0}   {1}".format(e,f), emid, flux):
        print >> fh, l
    if outfile != None:
        fh.close()
        if verbose: print ("Created: {0}".format(outfile))

def plot_chart_spectrum(**kwargs):
    """Plots the spectrum for input to ChaRT.

    The arguments are passed to get_chart_spectrum(); see that routine
    for more information.
    """

    vals = _get_chart_spectrum(**kwargs)
    c.open_undo_buffer()
    try:
        if c.info() == None:
            c.add_window(["id", "chart"])
        else:
            c.erase()

        c.add_curve(vals["x"], vals["y"], ["symbol.style","none"])
        c.set_plot_xlabel("Energy (keV)")
        c.set_plot_ylabel("Flux (photon cm^{-2} s^{-1})")
        c.set_plot_title("ChaRT Source Spectrum")

        id = c.ChipsId()
        id.coord_sys = c.FRAME_NORM
        c.add_label(id, 0.05, 0.97,"Model: {0}".format(vals["model"]))
        c.add_label(id, 0.95, 0.97,"Dataset: {0}".format(repr(vals["id"])),
                    ["halign", 1])

    except:
        # I really should only catch ChIPS errors here
        c.discard_undo_buffer()
        raise

    c.close_undo_buffer()

